from collections import defaultdict
from contextlib import contextmanager
import itertools
import os
import tempfile
import uuid

from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    PipelineDefinition,
    SolidInstance,
    check,
    execute_pipeline,
)

from dagster.core.execution import (
    ExecutionContext,
    RunConfig,
    construct_pipeline_execution_context,
    create_environment_config,
)

from dagster.core.utility_solids import define_stub_solid


def create_test_pipeline_execution_context(loggers=None, resources=None, tags=None):
    run_id = str(uuid.uuid4())
    pipeline_def = PipelineDefinition(name='test_legacy_context', solids=[])
    return construct_pipeline_execution_context(
        run_config=RunConfig(run_id, tags=tags, loggers=loggers),
        pipeline=pipeline_def,
        execution_context=ExecutionContext(),
        resources=resources,
        environment_config=create_environment_config(pipeline_def),
    )


def _unlink_swallow_errors(path):
    check.str_param(path, 'path')
    try:
        os.unlink(path)
    except:  # pylint: disable=W0702
        pass


@contextmanager
def get_temp_file_name():
    temp_file_name = tempfile.mkstemp()[1]
    try:
        yield temp_file_name
    finally:
        _unlink_swallow_errors(temp_file_name)


@contextmanager
def get_temp_file_names(number):
    check.int_param(number, 'number')

    temp_file_names = list()
    for _ in itertools.repeat(None, number):
        temp_file_name = tempfile.mkstemp()[1]
        temp_file_names.append(temp_file_name)

    try:
        yield tuple(temp_file_names)
    finally:
        for temp_file_name in temp_file_names:
            _unlink_swallow_errors(temp_file_name)


def _dep_key_of(solid):
    return SolidInstance(solid.definition.name, solid.name)


def build_pipeline_with_input_stubs(pipeline_def, inputs):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(inputs, 'inputs', key_type=str, value_type=dict)

    deps = defaultdict(dict)
    for solid_name, dep_dict in pipeline_def.dependencies.items():
        for input_name, dep in dep_dict.items():
            deps[solid_name][input_name] = dep

    stub_solid_defs = []

    for solid_name, input_dict in inputs.items():
        if not pipeline_def.has_solid(solid_name):
            raise DagsterInvariantViolationError(
                (
                    'You are injecting an input value for solid {solid_name} '
                    'into pipeline {pipeline_name} but that solid was not found'
                ).format(solid_name=solid_name, pipeline_name=pipeline_def.name)
            )

        solid = pipeline_def.solid_named(solid_name)
        for input_name, input_value in input_dict.items():
            stub_solid_def = define_stub_solid(
                '__stub_{solid_name}_{input_name}'.format(
                    solid_name=solid_name, input_name=input_name
                ),
                input_value,
            )
            stub_solid_defs.append(stub_solid_def)
            deps[_dep_key_of(solid)][input_name] = DependencyDefinition(stub_solid_def.name)

    return PipelineDefinition(
        name=pipeline_def.name + '_stubbed',
        solids=pipeline_def.solid_defs + stub_solid_defs,
        context_definitions=pipeline_def.context_definitions,
        dependencies=deps,
    )


def execute_solids(pipeline_def, solid_names, inputs=None, environment_dict=None):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.list_param(solid_names, 'solid_names', of_type=str)
    inputs = check.opt_dict_param(inputs, 'inputs', key_type=str, value_type=dict)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')

    sub_pipeline = pipeline_def.build_sub_pipeline(solid_names)
    stubbed_pipeline = build_pipeline_with_input_stubs(sub_pipeline, inputs)
    result = execute_pipeline(stubbed_pipeline, environment_dict)

    return {sr.solid.name: sr for sr in result.solid_result_list}


def execute_solid(pipeline_def, solid_name, inputs=None, environment_dict=None):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.str_param(solid_name, 'solid_name')
    inputs = check.opt_dict_param(inputs, 'inputs', key_type=str)
    environment_dict = check.opt_dict_param(environment_dict, 'environment')

    return execute_solids(
        pipeline_def, [solid_name], {solid_name: inputs} if inputs else None, environment_dict
    )[solid_name]



from collections import namedtuple

from dagster import check
from dagster.core.errors import (
    DagsterExecutionStepNotFoundError,
    DagsterInvalidSubplanInputNotFoundError,
    DagsterInvalidSubplanOutputNotFoundError,
    DagsterInvalidSubplanMissingInputError,
)

from .definitions import PipelineDefinition
from .execution_plan.intermediates_manager import StepOutputHandle

from .execution import (
    check_run_config_param,
    yield_pipeline_execution_context,
    create_execution_plan_core,
    invoke_executor_on_plan,
)


from .execution_context import InProcessExecutorConfig


class MarshalledOutput(namedtuple('_MarshalledOutput', 'output_name marshalling_key')):
    def __new__(cls, output_name, marshalling_key):
        return super(MarshalledOutput, cls).__new__(
            cls,
            check.str_param(output_name, 'output_name'),
            check.str_param(marshalling_key, 'marshalling_key'),
        )


MarshalledInput = namedtuple('MarshalledInput', 'input_name key')


class StepExecution(namedtuple('_StepExecution', 'step_key marshalled_inputs marshalled_outputs')):
    def __new__(cls, step_key, marshalled_inputs, marshalled_outputs):
        return super(StepExecution, cls).__new__(
            cls,
            check.str_param(step_key, 'step_key'),
            check.list_param(marshalled_inputs, 'marshalled_inputs', of_type=MarshalledInput),
            check.list_param(marshalled_outputs, 'marshalled_outputs', of_type=MarshalledOutput),
        )


def execute_marshalling(
    pipeline,
    step_keys,
    inputs_to_marshal=None,
    outputs_to_marshal=None,
    environment_dict=None,
    run_config=None,
):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.list_param(step_keys, 'step_keys', of_type=str)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')
    run_config = check_run_config_param(run_config)

    check.param_invariant(
        isinstance(run_config.executor_config, InProcessExecutorConfig),
        'run_config',
        (
            'In order for this function to work properly you must provide an executor config: '
            'for now, you must use the InProcessExecutorConfig'
        ),
    )

    intermediates_manager = run_config.executor_config.inmem_intermediates_manager

    with yield_pipeline_execution_context(
        pipeline, environment_dict, run_config
    ) as pipeline_context:

        execution_plan = create_execution_plan_core(pipeline_context)

        do_error_checking(
            inputs_to_marshal, execution_plan, pipeline_context, step_keys, outputs_to_marshal
        )

        unmarshal_inputs(inputs_to_marshal, execution_plan, pipeline_context, intermediates_manager)

        for step_key in step_keys:
            step = execution_plan.get_step_by_key(step_key)
            for step_input in step.step_inputs:
                if step_input.prev_output_handle.step_key in step_keys:
                    # dep in subset, we're fine
                    continue

                if intermediates_manager.has_value(step_input.prev_output_handle):
                    # dep preset in intermediates manager
                    continue

                raise DagsterInvalidSubplanMissingInputError(
                    (
                        'You have specified a subset execution on pipeline {pipeline_name} '
                        'with step_keys {step_keys}. You have failed to provide the required input '
                        '{input_name} for step {step_key}.'
                    ).format(
                        pipeline_name=pipeline_context.pipeline_def.name,
                        step_keys=step_keys,
                        input_name=step_input.name,
                        step_key=step.key,
                    ),
                    pipeline_name=pipeline_context.pipeline_def.name,
                    step_keys=step_keys,
                    input_name=step_input.name,
                    step=step,
                )

        execution_plan = create_execution_plan_core(pipeline_context)

        events = list(
            invoke_executor_on_plan(
                pipeline_context, execution_plan=execution_plan, step_keys_to_execute=step_keys
            )
        )

        marshal_outputs(
            events, outputs_to_marshal, execution_plan, intermediates_manager, pipeline_context
        )

        return events


def do_error_checking(
    inputs_to_marshal, execution_plan, pipeline_context, step_keys, outputs_to_marshal
):
    if inputs_to_marshal:
        for step_key, input_dict in inputs_to_marshal.items():
            for input_name, _marshalling_key in input_dict.items():
                if not execution_plan.has_step(step_key):
                    raise DagsterExecutionStepNotFoundError(step_key=step_key)
                step = execution_plan.get_step_by_key(step_key)
                if not step.has_step_input(input_name):
                    raise DagsterInvalidSubplanInputNotFoundError(
                        'Input {input_name} on {step_key} does not exist.'.format(
                            input_name=input_name, step_key=step_key
                        ),
                        pipeline_name=pipeline_context.pipeline_def.name,
                        step_keys=step_keys,
                        input_name=input_name,
                        step=step,
                    )

    if outputs_to_marshal:
        for step_key, marshalled_outputs in outputs_to_marshal.items():
            if not execution_plan.has_step(step_key):
                raise DagsterExecutionStepNotFoundError(step_key=step_key)
            step = execution_plan.get_step_by_key(step_key)
            for marshalled_output in marshalled_outputs:
                if not step.has_step_output(marshalled_output.output_name):
                    raise DagsterInvalidSubplanOutputNotFoundError(
                        'Execution step {step_key} does not have output {output}'.format(
                            step_key=step_key, output=marshalled_output.output_name
                        ),
                        pipeline_name=pipeline_context.pipeline_def.name,
                        step_keys=step_keys,
                        step=step,
                        output_name=marshalled_output.output_name,
                    )


def marshal_outputs(
    events, outputs_to_marshal, execution_plan, intermediates_manager, pipeline_context
):
    if outputs_to_marshal:
        successful_outputs = set()
        for event in events:
            if event.is_successful_output:
                successful_outputs.add(
                    StepOutputHandle(event.step_key, event.step_output_data.output_name)
                )

        for step_key, marshalled_outputs in outputs_to_marshal.items():
            if not execution_plan.has_step(step_key):
                raise DagsterExecutionStepNotFoundError(step_key=step_key)
            step = execution_plan.get_step_by_key(step_key)
            for marshalled_output in marshalled_outputs:
                step_output = step.step_output_named(marshalled_output.output_name)
                step_output_handle = StepOutputHandle(step_key, marshalled_output.output_name)

                if step_output_handle not in successful_outputs:
                    continue

                output_value = intermediates_manager.get_value(step_output_handle)
                pipeline_context.persistence_strategy.write_value(
                    step_output.runtime_type.serialization_strategy,
                    marshalled_output.marshalling_key,
                    output_value,
                )


def unmarshal_inputs(inputs_to_marshal, execution_plan, pipeline_context, intermediates_manager):
    if inputs_to_marshal:
        for step_key, input_dict in inputs_to_marshal.items():
            for input_name, marshalling_key in input_dict.items():
                step = execution_plan.get_step_by_key(step_key)
                step_input = step.step_input_named(input_name)
                step_output_handle = step_input.prev_output_handle
                input_value = pipeline_context.persistence_strategy.read_value(
                    step_input.runtime_type.serialization_strategy, marshalling_key
                )
                intermediates_manager.set_value(step_output_handle, input_value)



from .context import PipelineContextDefinition

from .dependency import (
    DependencyDefinition,
    Solid,
    SolidInputHandle,
    SolidInstance,
    SolidOutputHandle,
)

from .expectation import ExpectationDefinition, ExpectationResult

from .input import InputDefinition

from .output import OutputDefinition

from .resource import ResourceDefinition

from .result import Result

from .repository import RepositoryDefinition

from .pipeline import PipelineDefinition, solids_in_topological_order

from .pipeline_creation import create_execution_structure

from .solid import SolidDefinition



from collections import namedtuple

from dagster import check

DEFAULT_CONTEXT_NAME = 'default'


def _default_persistence_config():
    return {'file': {}}


# lifted from https://bit.ly/2HcQAuv
class ContextConfig(namedtuple('_ContextConfig', 'name config resources persistence')):
    def __new__(cls, name=None, config=None, resources=None, persistence=None):
        return super(ContextConfig, cls).__new__(
            cls,
            check.opt_str_param(name, 'name', DEFAULT_CONTEXT_NAME),
            config,
            check.opt_dict_param(resources, 'resources', key_type=str),
            _default_persistence_config()
            if persistence is None
            else check.dict_param(persistence, 'persistence', key_type=str),
        )


class SolidConfig(namedtuple('_SolidConfig', 'config inputs outputs')):
    def __new__(cls, config, inputs=None, outputs=None):
        return super(SolidConfig, cls).__new__(
            cls,
            config,
            check.opt_dict_param(inputs, 'inputs', key_type=str),
            check.opt_list_param(outputs, 'outputs', of_type=dict),
        )


class EnvironmentConfig(
    namedtuple('_EnvironmentConfig', 'context solids expectations execution original_config_dict')
):
    def __new__(
        cls, solids=None, context=None, expectations=None, execution=None, original_config_dict=None
    ):
        check.opt_inst_param(context, 'context', ContextConfig)
        check.opt_inst_param(expectations, 'expectations', ExpectationsConfig)
        check.opt_inst_param(execution, 'execution', ExecutionConfig)
        check.opt_dict_param(original_config_dict, 'original_config_dict')

        if context is None:
            context = ContextConfig()

        if expectations is None:
            expectations = ExpectationsConfig(evaluate=True)

        if execution is None:
            execution = ExecutionConfig()

        return super(EnvironmentConfig, cls).__new__(
            cls,
            context=context,
            solids=check.opt_dict_param(solids, 'solids', key_type=str, value_type=SolidConfig),
            expectations=expectations,
            execution=execution,
            original_config_dict=original_config_dict,
        )


class ExpectationsConfig(namedtuple('_ExpecationsConfig', 'evaluate')):
    def __new__(cls, evaluate):
        return super(ExpectationsConfig, cls).__new__(
            cls, evaluate=check.bool_param(evaluate, 'evaluate')
        )


class ExecutionConfig(namedtuple('_ExecutionConfig', '')):
    def __new__(cls):

        return super(ExecutionConfig, cls).__new__(cls)



import pytest

from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineConfigEvaluationError,
    PipelineDefinition,
    SolidInstance,
    execute_pipeline,
    solid,
    types,
)


def test_string_from_inputs():
    called = {}

    @solid(inputs=[InputDefinition('string_input', types.String)])
    def str_as_input(_context, string_input):
        assert string_input == 'foo'
        called['yup'] = True

    pipeline = PipelineDefinition(name='test_string_from_inputs_pipeline', solids=[str_as_input])

    result = execute_pipeline(
        pipeline, {'solids': {'str_as_input': {'inputs': {'string_input': {'value': 'foo'}}}}}
    )

    assert result.success
    assert called['yup']


def test_string_from_aliased_inputs():
    called = {}

    @solid(inputs=[InputDefinition('string_input', types.String)])
    def str_as_input(_context, string_input):
        assert string_input == 'foo'
        called['yup'] = True

    pipeline = PipelineDefinition(
        solids=[str_as_input], dependencies={SolidInstance('str_as_input', alias='aliased'): {}}
    )

    result = execute_pipeline(
        pipeline, {'solids': {'aliased': {'inputs': {'string_input': {'value': 'foo'}}}}}
    )

    assert result.success
    assert called['yup']


def test_string_missing_inputs():
    called = {}

    @solid(inputs=[InputDefinition('string_input', types.String)])
    def str_as_input(_context, string_input):  # pylint: disable=W0613
        called['yup'] = True

    pipeline = PipelineDefinition(name='missing_inputs', solids=[str_as_input])
    with pytest.raises(PipelineConfigEvaluationError) as exc_info:
        execute_pipeline(pipeline)

    assert (
        'Missing required field  "solids" at document config root. Expected: "{ '
        'context?: MissingInputs.ContextConfig execution?: '
        'MissingInputs.ExecutionConfig expectations?: MissingInputs.ExpectationsConfig '
        'solids: MissingInputs.SolidsConfigDictionary }"'
    ) in str(exc_info.value)

    assert 'yup' not in called


def test_string_missing_input_collision():
    called = {}

    @solid(outputs=[OutputDefinition(types.String)])
    def str_as_output(_context):
        return 'bar'

    @solid(inputs=[InputDefinition('string_input', types.String)])
    def str_as_input(_context, string_input):  # pylint: disable=W0613
        called['yup'] = True

    pipeline = PipelineDefinition(
        name='overlapping',
        solids=[str_as_input, str_as_output],
        dependencies={'str_as_input': {'string_input': DependencyDefinition('str_as_output')}},
    )
    with pytest.raises(PipelineConfigEvaluationError) as exc_info:
        execute_pipeline(
            pipeline, {'solids': {'str_as_input': {'inputs': {'string_input': 'bar'}}}}
        )

    assert 'Error 1: Undefined field "inputs" at path root:solids:str_as_input' in str(
        exc_info.value
    )

    assert 'yup' not in called



# pylint: disable=W0622,W0614,W0401
import pytest


from dagster import DagsterInvariantViolationError, execute_pipeline
from dagster.tutorials.intro_tutorial.multiple_outputs import (
    define_multiple_outputs_step_one_pipeline,
    define_multiple_outputs_step_two_pipeline,
    define_multiple_outputs_step_three_pipeline,
)


def test_intro_tutorial_multiple_outputs_step_one():
    result = execute_pipeline(define_multiple_outputs_step_one_pipeline())

    assert result.success
    assert result.result_for_solid('return_dict_results').transformed_value('out_one') == 23
    assert result.result_for_solid('return_dict_results').transformed_value('out_two') == 45
    assert result.result_for_solid('log_num').transformed_value() == 23
    assert result.result_for_solid('log_num_squared').transformed_value() == 45 * 45


def test_intro_tutorial_multiple_outputs_step_two():
    result = execute_pipeline(define_multiple_outputs_step_two_pipeline())

    assert result.success
    assert result.result_for_solid('yield_outputs').transformed_value('out_one') == 23
    assert result.result_for_solid('yield_outputs').transformed_value('out_two') == 45
    assert result.result_for_solid('log_num').transformed_value() == 23
    assert result.result_for_solid('log_num_squared').transformed_value() == 45 * 45


def test_intro_tutorial_multiple_outputs_step_three():
    result = execute_pipeline(
        define_multiple_outputs_step_three_pipeline(),
        {'solids': {'conditional': {'config': 'out_two'}}},
    )

    # successful things
    assert result.success
    assert result.result_for_solid('conditional').transformed_value('out_two') == 45
    assert result.result_for_solid('log_num_squared').transformed_value() == 45 * 45

    # unsuccessful things
    with pytest.raises(DagsterInvariantViolationError):
        assert result.result_for_solid('conditional').transformed_value('out_one') == 45


if __name__ == '__main__':
    execute_pipeline(
        define_multiple_outputs_step_three_pipeline(),
        {'solids': {'conditional': {'config': 'out_two'}}},
    )

    # execute_pipeline(define_multiple_outputs_step_two())



'''
Naming conventions:

For public functions:

execute_*

These represent functions which do purely in-memory compute. They will evaluate expectations
the core transform, and exercise all logging and metrics tracking (outside of outputs), but they
will not invoke *any* outputs (and their APIs don't allow the user to).


'''

# too many lines
# pylint: disable=C0302

from collections import defaultdict, OrderedDict
from contextlib import contextmanager
import inspect
import itertools

from contextlib2 import ExitStack
from dagster import check
from dagster.utils import merge_dicts

from .definitions import PipelineDefinition, Solid
from .definitions.utils import DEFAULT_OUTPUT
from .definitions.environment_configs import construct_environment_config

from .execution_context import (
    RunConfig,
    InProcessExecutorConfig,
    MultiprocessExecutorConfig,
    SystemPipelineExecutionContextData,
    SystemPipelineExecutionContext,
)


from .errors import DagsterInvariantViolationError

from .events import construct_event_logger

from .execution_plan.create import create_execution_plan_core

from .execution_plan.objects import (
    ExecutionPlan,
    ExecutionStepEvent,
    ExecutionStepEventType,
    StepKind,
)

from .execution_plan.multiprocessing_engine import multiprocess_execute_plan

from .execution_plan.simple_engine import start_inprocess_executor

from .files import LocalTempFileStore

from .init_context import InitContext, InitResourceContext

from .log import DagsterLog

from .system_config.objects import EnvironmentConfig

from .types.evaluator import EvaluationError, evaluate_config_value, friendly_string_for_error
from .types.marshal import FilePersistencePolicy

from .user_context import ExecutionContext


class PipelineExecutionResult(object):
    '''Result of execution of the whole pipeline. Returned eg by :py:func:`execute_pipeline`.

    Attributes:
        pipeline (PipelineDefinition): Pipeline that was executed
        context (ExecutionContext): ExecutionContext of that particular Pipeline run.
        result_list (list[SolidExecutionResult]): List of results for each pipeline solid.
    '''

    def __init__(self, pipeline, run_id, step_event_list):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.run_id = check.str_param(run_id, 'run_id')
        self.step_event_list = check.list_param(
            step_event_list, 'step_event_list', of_type=ExecutionStepEvent
        )

        solid_result_dict = self._context_solid_result_dict(step_event_list)

        self.solid_result_dict = solid_result_dict
        self.solid_result_list = list(self.solid_result_dict.values())

    def _context_solid_result_dict(self, step_event_list):
        solid_set = set()
        solid_order = []
        step_events_by_solid_by_kind = defaultdict(lambda: defaultdict(list))

        for step_event in step_event_list:
            solid_name = step_event.solid_name
            if solid_name not in solid_set:
                solid_order.append(solid_name)
                solid_set.add(solid_name)

            step_events_by_solid_by_kind[solid_name][step_event.step_kind].append(step_event)

        solid_result_dict = OrderedDict()

        for solid_name in solid_order:
            solid_result_dict[solid_name] = SolidExecutionResult(
                self.pipeline.solid_named(solid_name),
                dict(step_events_by_solid_by_kind[solid_name]),
            )
        return solid_result_dict

    @property
    def success(self):
        '''Whether the pipeline execution was successful at all steps'''
        return all([not step_event.is_step_failure for step_event in self.step_event_list])

    def result_for_solid(self, name):
        '''Get a :py:class:`SolidExecutionResult` for a given solid name.

        Returns:
          SolidExecutionResult
        '''
        check.str_param(name, 'name')

        if not self.pipeline.has_solid(name):
            raise DagsterInvariantViolationError(
                'Try to get result for solid {name} in {pipeline}. No such solid.'.format(
                    name=name, pipeline=self.pipeline.display_name
                )
            )

        if name not in self.solid_result_dict:
            raise DagsterInvariantViolationError(
                'Did not find result for solid {name} in pipeline execution result'.format(
                    name=name
                )
            )

        return self.solid_result_dict[name]


class SolidExecutionResult(object):
    '''Execution result for one solid of the pipeline.

    Attributes:
      context (ExecutionContext): ExecutionContext of that particular Pipeline run.
      solid (SolidDefinition): Solid for which this result is
    '''

    def __init__(self, solid, step_events_by_kind):
        self.solid = check.inst_param(solid, 'solid', Solid)
        self.step_events_by_kind = check.dict_param(
            step_events_by_kind, 'step_events_by_kind', key_type=StepKind, value_type=list
        )

    @property
    def transform(self):
        check.invariant(len(self.step_events_by_kind[StepKind.TRANSFORM]) == 1)
        return self.step_events_by_kind[StepKind.TRANSFORM][0]

    @property
    def transforms(self):
        return self.step_events_by_kind.get(StepKind.TRANSFORM, [])

    @property
    def input_expectations(self):
        return self.step_events_by_kind.get(StepKind.INPUT_EXPECTATION, [])

    @property
    def output_expectations(self):
        return self.step_events_by_kind.get(StepKind.OUTPUT_EXPECTATION, [])

    @staticmethod
    def from_step_events(pipeline_context, step_events):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        step_events = check.list_param(step_events, 'step_events', ExecutionStepEvent)
        if step_events:
            step_events_by_kind = defaultdict(list)

            solid = None
            for result in step_events:
                if solid is None:
                    solid = result.step.solid
                check.invariant(result.step.solid is solid, 'Must all be from same solid')

            for result in step_events:
                step_events_by_kind[result.kind].append(result)

            return SolidExecutionResult(
                solid=step_events[0].step.solid, step_events_by_kind=dict(step_events_by_kind)
            )
        else:
            check.failed("Cannot create SolidExecutionResult from empty list")

    @property
    def success(self):
        '''Whether the solid execution was successful'''
        return all(
            [
                not step_event.is_step_failure
                for step_event in itertools.chain(
                    self.input_expectations, self.output_expectations, self.transforms
                )
            ]
        )

    @property
    def transformed_values(self):
        '''Return dictionary of transformed results, with keys being output names.
        Returns None if execution result isn't a success.'''
        if self.success and self.transforms:
            return {
                result.step_output_data.output_name: result.step_output_data.get_value()
                for result in self.transforms
            }
        else:
            return None

    def transformed_value(self, output_name=DEFAULT_OUTPUT):
        '''Returns transformed value either for DEFAULT_OUTPUT or for the output
        given as output_name. Returns None if execution result isn't a success'''
        check.str_param(output_name, 'output_name')

        if not self.solid.definition.has_output(output_name):
            raise DagsterInvariantViolationError(
                '{output_name} not defined in solid {solid}'.format(
                    output_name=output_name, solid=self.solid.name
                )
            )

        if self.success:
            for result in self.transforms:
                if result.step_output_data.output_name == output_name:
                    return result.step_output_data.get_value()
            raise DagsterInvariantViolationError(
                (
                    'Did not find result {output_name} in solid {self.solid.name} '
                    'execution result'
                ).format(output_name=output_name, self=self)
            )
        else:
            return None

    @property
    def failure_data(self):
        '''Returns the failing step's data that happened during this solid's execution, if any'''
        for result in itertools.chain(
            self.input_expectations, self.output_expectations, self.transforms
        ):
            if result.event_type == ExecutionStepEventType.STEP_FAILURE:
                return result.step_failure_data


def check_run_config_param(run_config):
    return (
        check.inst_param(run_config, 'run_config', RunConfig)
        if run_config
        else RunConfig(executor_config=InProcessExecutorConfig())
    )


def create_execution_plan(pipeline, environment_dict=None, run_config=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict', key_type=str)
    run_config = check_run_config_param(run_config)
    check.inst_param(run_config, 'run_config', RunConfig)

    with yield_pipeline_execution_context(
        pipeline, environment_dict, run_config
    ) as pipeline_context:
        return create_execution_plan_core(pipeline_context)


def get_tags(user_context_params, run_config, pipeline):
    check.inst_param(user_context_params, 'user_context_params', ExecutionContext)
    check.opt_inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    base_tags = merge_dicts({'pipeline': pipeline.name}, user_context_params.tags)

    if run_config and run_config.tags:
        user_keys = set(user_context_params.tags.keys())
        provided_keys = set(run_config.tags.keys())
        if not user_keys.isdisjoint(provided_keys):
            raise DagsterInvariantViolationError(
                (
                    'You have specified tags and user-defined tags '
                    'that overlap. User keys: {user_keys}. Reentrant keys: '
                    '{provided_keys}.'
                ).format(user_keys=user_keys, provided_keys=provided_keys)
            )

        return merge_dicts(base_tags, run_config.tags)
    else:
        return base_tags


def _ensure_gen(thing_or_gen):
    if not inspect.isgenerator(thing_or_gen):

        def _gen_thing():
            yield thing_or_gen

        return _gen_thing()

    return thing_or_gen


@contextmanager
def with_maybe_gen(thing_or_gen):
    gen = _ensure_gen(thing_or_gen)

    try:
        thing = next(gen)
    except StopIteration:
        check.failed('Must yield one item. You did not yield anything.')

    yield thing

    stopped = False

    try:
        next(gen)
    except StopIteration:
        stopped = True

    check.invariant(stopped, 'Must yield one item. Yielded more than one item')


def _create_persistence_strategy(persistence_config):
    check.dict_param(persistence_config, 'persistence_config', key_type=str)

    persistence_key, _config_value = list(persistence_config.items())[0]

    if persistence_key == 'file':
        return FilePersistencePolicy()
    else:
        check.failed('Unsupported persistence key: {}'.format(persistence_key))


@contextmanager
def yield_pipeline_execution_context(pipeline_def, environment_dict, run_config):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(environment_dict, 'environment_dict', key_type=str)
    check.inst_param(run_config, 'run_config', RunConfig)

    environment_config = create_environment_config(pipeline_def, environment_dict)

    context_definition = pipeline_def.context_definitions[environment_config.context.name]

    ec_or_gen = context_definition.context_fn(
        InitContext(
            context_config=environment_config.context.config,
            pipeline_def=pipeline_def,
            run_id=run_config.run_id,
        )
    )

    with with_maybe_gen(ec_or_gen) as execution_context:
        check.inst(execution_context, ExecutionContext)

        with _create_resources(
            pipeline_def,
            context_definition,
            environment_config,
            execution_context,
            run_config.run_id,
        ) as resources:
            yield construct_pipeline_execution_context(
                run_config, execution_context, pipeline_def, resources, environment_config
            )


def construct_pipeline_execution_context(
    run_config, execution_context, pipeline, resources, environment_config
):
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(execution_context, 'execution_context', ExecutionContext)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)

    loggers = _create_loggers(run_config, execution_context)
    tags = get_tags(execution_context, run_config, pipeline)
    log = DagsterLog(run_config.run_id, tags, loggers)

    return SystemPipelineExecutionContext(
        SystemPipelineExecutionContextData(
            pipeline_def=pipeline,
            run_config=run_config,
            resources=resources,
            environment_config=environment_config,
            persistence_strategy=_create_persistence_strategy(
                environment_config.context.persistence
            ),
            files=LocalTempFileStore(run_config.run_id),
        ),
        tags=tags,
        log=log,
    )


def _create_loggers(run_config, execution_context):
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(execution_context, 'execution_context', ExecutionContext)

    if run_config.event_callback:
        return execution_context.loggers + [construct_event_logger(run_config.event_callback)]
    elif run_config.loggers:
        return execution_context.loggers + run_config.loggers
    else:
        return execution_context.loggers


@contextmanager
def _create_resources(pipeline_def, context_def, environment, execution_context, run_id):
    if not context_def.resources:
        yield execution_context.resources
        return

    resources = {}
    check.invariant(
        not execution_context.resources,
        (
            'If resources explicitly specified on context definition, the context '
            'creation function should not return resources as a property of the '
            'ExecutionContext.'
        ),
    )

    # See https://bit.ly/2zIXyqw
    # The "ExitStack" allows one to stack up N context managers and then yield
    # something. We do this so that resources can cleanup after themselves. We
    # can potentially have many resources so we need to use this abstraction.
    with ExitStack() as stack:
        for resource_name in context_def.resources.keys():
            resource_obj_or_gen = get_resource_or_gen(
                pipeline_def, context_def, resource_name, environment, run_id
            )

            resource_obj = stack.enter_context(with_maybe_gen(resource_obj_or_gen))

            resources[resource_name] = resource_obj

        context_name = environment.context.name

        resources_type = pipeline_def.context_definitions[context_name].resources_type
        yield resources_type(**resources)


def get_resource_or_gen(pipeline_def, context_definition, resource_name, environment, run_id):
    resource_def = context_definition.resources[resource_name]
    # Need to do default values
    resource_config = environment.context.resources.get(resource_name, {}).get('config')
    return resource_def.resource_fn(
        InitResourceContext(
            pipeline_def=pipeline_def,
            resource_def=resource_def,
            context_config=environment.context.config,
            resource_config=resource_config,
            run_id=run_id,
        )
    )


def execute_pipeline_iterator(pipeline, environment_dict=None, run_config=None):
    '''Returns iterator that yields :py:class:`SolidExecutionResult` for each
    solid executed in the pipeline.

    This is intended to allow the caller to do things between each executed
    node. For the 'synchronous' API, see :py:func:`execute_pipeline`.

    Parameters:
      pipeline (PipelineDefinition): pipeline to run
      execution (ExecutionContext): execution context of the run
    '''
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')
    run_config = check_run_config_param(run_config)

    with yield_pipeline_execution_context(
        pipeline, environment_dict, run_config
    ) as pipeline_context:

        pipeline_context.events.pipeline_start()

        execution_plan = create_execution_plan_core(pipeline_context)

        steps = execution_plan.topological_steps()

        if not steps:
            pipeline_context.log.debug(
                'Pipeline {pipeline} has no nodes and no execution will happen'.format(
                    pipeline=pipeline.display_name
                )
            )
            pipeline_context.events.pipeline_success()
            return

        pipeline_context.log.debug(
            'About to execute the compute node graph in the following order {order}'.format(
                order=[step.key for step in steps]
            )
        )

        check.invariant(len(steps[0].step_inputs) == 0)

        pipeline_success = True

        for step_event in invoke_executor_on_plan(pipeline_context, execution_plan):
            if step_event.is_step_failure:
                pipeline_success = False
            yield step_event

        if pipeline_success:
            pipeline_context.events.pipeline_success()
        else:
            pipeline_context.events.pipeline_failure()


def execute_pipeline(pipeline, environment_dict=None, run_config=None):
    '''
    "Synchronous" version of :py:func:`execute_pipeline_iterator`.

    Note: throw_on_user_error is very useful in testing contexts when not testing for error
    conditions

    Parameters:
      pipeline (PipelineDefinition): Pipeline to run
      environment (dict): The enviroment that parameterizes this run

    Returns:
      PipelineExecutionResult
    '''

    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')
    run_config = check_run_config_param(run_config)

    return PipelineExecutionResult(
        pipeline,
        run_config.run_id,
        list(
            execute_pipeline_iterator(
                pipeline=pipeline, environment_dict=environment_dict, run_config=run_config
            )
        ),
    )


class PipelineConfigEvaluationError(Exception):
    def __init__(self, pipeline, errors, config_value, *args, **kwargs):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.errors = check.list_param(errors, 'errors', of_type=EvaluationError)
        self.config_value = config_value

        error_msg = 'Pipeline "{pipeline}" config errors:'.format(pipeline=pipeline.name)

        error_messages = []

        for i_error, error in enumerate(self.errors):
            error_message = friendly_string_for_error(error)
            error_messages.append(error_message)
            error_msg += '\n    Error {i_error}: {error_message}'.format(
                i_error=i_error + 1, error_message=error_message
            )

        self.message = error_msg
        self.error_messages = error_messages

        super(PipelineConfigEvaluationError, self).__init__(error_msg, *args, **kwargs)


def invoke_executor_on_plan(pipeline_context, execution_plan, step_keys_to_execute=None):
    if isinstance(pipeline_context.executor_config, InProcessExecutorConfig):
        step_events_gen = start_inprocess_executor(
            pipeline_context,
            execution_plan,
            pipeline_context.run_config.executor_config.inmem_intermediates_manager,
            step_keys_to_execute,
        )
    elif isinstance(pipeline_context.executor_config, MultiprocessExecutorConfig):
        check.invariant(not step_keys_to_execute, 'subplan not supported for multiprocess yet')
        step_events_gen = multiprocess_execute_plan(pipeline_context, execution_plan)
    else:
        check.failed('Unsupported config {}'.format(pipeline_context.executor_config))

    for step_event in step_events_gen:
        yield step_event


def execute_plan(execution_plan, environment_dict=None, run_config=None):
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')
    run_config = check_run_config_param(run_config)

    with yield_pipeline_execution_context(
        execution_plan.pipeline_def, environment_dict, run_config
    ) as pipeline_context:
        return list(invoke_executor_on_plan(pipeline_context, execution_plan))


def create_environment_config(pipeline, environment_dict=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(environment_dict, 'environment')

    result = evaluate_config_value(pipeline.environment_type, environment_dict)

    if not result.success:
        raise PipelineConfigEvaluationError(pipeline, result.errors, environment_dict)

    return construct_environment_config(result.value)


class ExecutionSelector(object):
    def __init__(self, name, solid_subset=None):
        self.name = check.str_param(name, 'name')
        if solid_subset is None:
            self.solid_subset = None
        else:
            self.solid_subset = check.opt_list_param(solid_subset, 'solid_subset', of_type=str)



from contextlib import contextmanager
import sys

from future.utils import raise_from
import six

from dagster import check

from dagster.utils.logging import get_formatted_stack_trace

from dagster.utils.timing import time_execution_scope

from dagster.core.errors import (
    DagsterError,
    DagsterInvariantViolationError,
    DagsterRuntimeCoercionError,
    DagsterExecutionStepExecutionError,
    DagsterTypeError,
)

from dagster.core.execution_context import (
    ExecutorConfig,
    SystemPipelineExecutionContext,
    SystemStepExecutionContext,
)

from dagster.utils.error import serializable_error_info_from_exc_info

from .intermediates_manager import IntermediatesManager

from .objects import (
    ExecutionPlan,
    ExecutionStep,
    ExecutionStepEvent,
    StepOutputHandle,
    StepOutputValue,
    StepOutputData,
    StepFailureData,
)


def _all_inputs_covered(step, intermediates_manager):
    for step_input in step.step_inputs:
        if not intermediates_manager.has_value(step_input.prev_output_handle):
            return False
    return True


def start_inprocess_executor(
    pipeline_context, execution_plan, intermediates_manager, step_keys_to_execute=None
):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    step_key_set = None if step_keys_to_execute is None else set(step_keys_to_execute)

    check.param_invariant(
        isinstance(pipeline_context.executor_config, ExecutorConfig),
        'pipeline_context',
        'Expected executor_config to be ExecutorConfig got {}'.format(
            pipeline_context.executor_config
        ),
    )

    step_levels = execution_plan.topological_step_levels()

    # It would be good to implement a reference tracking algorithm here so we could
    # garbage collection results that are no longer needed by any steps
    # https://github.com/dagster-io/dagster/issues/811

    for step_level in step_levels:
        for step in step_level:
            if step_key_set and step.key not in step_key_set:
                continue

            step_context = pipeline_context.for_step(step)

            if not intermediates_manager.all_inputs_covered(step):
                expected_outputs = [ni.prev_output_handle for ni in step.step_inputs]

                step_context.log.info(
                    (
                        'Not all inputs covered for {step}. Not executing. Output need for '
                        'inputs {expected_outputs}'
                    ).format(expected_outputs=expected_outputs, step=step.key)
                )
                continue

            input_values = _create_input_values(step, intermediates_manager)

            for step_event in check.generator(
                execute_step_in_memory(step_context, input_values, intermediates_manager)
            ):
                check.inst(step_event, ExecutionStepEvent)
                if (
                    pipeline_context.executor_config.throw_on_user_error
                    and step_event.is_step_failure
                ):
                    step_event.reraise_user_error()

                yield step_event


def _create_input_values(step, intermediates_manager):
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    input_values = {}
    for step_input in step.step_inputs:
        prev_output_handle = step_input.prev_output_handle
        input_value = intermediates_manager.get_value(prev_output_handle)
        input_values[step_input.name] = input_value
    return input_values


def execute_step_in_memory(step_context, inputs, intermediates_manager):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    try:
        for step_event in check.generator(
            _execute_steps_core_loop(step_context, inputs, intermediates_manager)
        ):
            step_context.log.info(
                'Step {step} emitted {value} for output {output}'.format(
                    step=step_context.step.key,
                    value=step_event.step_output_data.value_repr,
                    output=step_event.step_output_data.output_name,
                )
            )
            yield step_event
    except DagsterError as dagster_error:
        step_context.log.error(str(dagster_error))

        exc_info = (
            # pylint does not know original_exc_info exists is is_user_code_error is true
            # pylint: disable=no-member
            dagster_error.original_exc_info
            if dagster_error.is_user_code_error
            else sys.exc_info()
        )

        if step_context.executor_config.throw_on_user_error:
            six.reraise(*exc_info)

        error_info = serializable_error_info_from_exc_info(exc_info)

        yield ExecutionStepEvent.step_failure_event(
            step_context=step_context,
            step_failure_data=StepFailureData(
                error_message=error_info.message,
                error_cls_name=exc_info[0].__name__,  # 0 is the exception type
                stack=error_info.stack,
            ),
        )
        return


def _error_check_step_output_values(step, step_output_values):
    check.inst_param(step, 'step', ExecutionStep)
    check.generator_param(step_output_values, 'step_output_values')

    seen_outputs = set()
    for step_output_value in step_output_values:
        if not step.has_step_output(step_output_value.output_name):
            output_names = list(
                [output_def.name for output_def in step.solid.definition.output_defs]
            )
            raise DagsterInvariantViolationError(
                'Core transform for {step.solid.name} returned an output '
                '{step_output_value.output_name} that does not exist. The available '
                'outputs are {output_names}'.format(
                    step=step, step_output_value=step_output_value, output_names=output_names
                )
            )

        if step_output_value.output_name in seen_outputs:
            raise DagsterInvariantViolationError(
                'Core transform for {step.solid.name} returned an output '
                '{step_output_value.output_name} multiple times'.format(
                    step=step, step_output_value=step_output_value
                )
            )

        yield step_output_value
        seen_outputs.add(step_output_value.output_name)


def _execute_steps_core_loop(step_context, inputs, intermediates_manager):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    evaluated_inputs = {}
    # do runtime type checks of inputs versus step inputs
    for input_name, input_value in inputs.items():
        evaluated_inputs[input_name] = _get_evaluated_input(
            step_context.step, input_name, input_value
        )

    step_output_value_iterator = check.generator(
        _iterate_step_output_values_within_boundary(step_context, evaluated_inputs)
    )

    for step_output_value in check.generator(
        _error_check_step_output_values(step_context.step, step_output_value_iterator)
    ):

        yield _create_step_event(step_context, step_output_value, intermediates_manager)


def _create_step_event(step_context, step_output_value, intermediates_manager):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.inst_param(step_output_value, 'step_output_value', StepOutputValue)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    step = step_context.step
    step_output = step.step_output_named(step_output_value.output_name)

    try:
        value = step_output.runtime_type.coerce_runtime_value(step_output_value.value)
        step_output_handle = StepOutputHandle.from_step(
            step=step, output_name=step_output_value.output_name
        )

        intermediates_manager.set_value(step_output_handle, value)

        return ExecutionStepEvent.step_output_event(
            step_context=step_context,
            step_output_data=StepOutputData(
                step_output_handle=step_output_handle,
                value_repr=repr(value),
                intermediates_manager=intermediates_manager,
            ),
        )
    except DagsterRuntimeCoercionError as e:
        raise DagsterInvariantViolationError(
            (
                'Solid {step.solid.name} output name {output_name} output '
                '{step_output_value.value} type failure: {error_msg}.'
            ).format(
                step=step,
                step_output_value=step_output_value,
                error_msg=','.join(e.args),
                output_name=step_output_value.output_name,
            )
        )


def _get_evaluated_input(step, input_name, input_value):
    check.inst_param(step, 'step', ExecutionStep)
    check.str_param(input_name, 'input_name')

    step_input = step.step_input_named(input_name)
    try:
        return step_input.runtime_type.coerce_runtime_value(input_value)
    except DagsterRuntimeCoercionError as evaluate_error:
        raise_from(
            DagsterTypeError(
                (
                    'Solid {step.solid.name} input {input_name} received value {input_value} '
                    'which does not pass the typecheck for Dagster type '
                    '{step_input.runtime_type.name}. Step {step.key}'
                ).format(
                    step=step, input_name=input_name, input_value=input_value, step_input=step_input
                )
            ),
            evaluate_error,
        )


def _iterate_step_output_values_within_boundary(step_context, evaluated_inputs):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.dict_param(evaluated_inputs, 'evaluated_inputs', key_type=str)

    error_str = 'Error occured during step {key}'.format(key=step_context.step.key)
    with _execution_step_error_boundary(step_context, error_str):
        gen = check.opt_generator(step_context.step.compute_fn(step_context, evaluated_inputs))

        if gen is not None:
            for step_output_value in gen:
                yield step_output_value


@contextmanager
def _execution_step_error_boundary(step_context, msg, **kwargs):
    '''
    Wraps the execution of user-space code in an error boundary. This places a uniform
    policy around an user code invoked by the framework. This ensures that all user
    errors are wrapped in the SolidUserCodeExecutionError, and that the original stack
    trace of the user error is preserved, so that it can be reported without confusing
    framework code in the stack trace, if a tool author wishes to do so. This has
    been especially help in a notebooking context.
    '''
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.str_param(msg, 'msg')

    step = step_context.step
    step_context.events.execution_plan_step_start(step.key)
    try:
        with time_execution_scope() as timer_result:
            yield

        step_context.events.execution_plan_step_success(step.key, timer_result.millis)
    except Exception as e:  # pylint: disable=W0703
        step_context.events.execution_plan_step_failure(step.key, sys.exc_info())

        stack_trace = get_formatted_stack_trace(e)
        step_context.log.error(str(e), stack_trace=stack_trace)

        if isinstance(e, DagsterError):
            raise e
        else:
            raise_from(
                DagsterExecutionStepExecutionError(
                    msg.format(**kwargs), user_exception=e, original_exc_info=sys.exc_info()
                ),
                e,
            )



import subprocess


def check_script(path):
    subprocess.check_output(['python', path])


def check_cli_execute_file_pipeline(path, pipeline_fn_name, env_file=None):
    cli_cmd = [
        'python',
        '-m',
        'dagster',
        'pipeline',
        'execute',
        '-f',
        path,
        '-n',
        pipeline_fn_name,
    ]

    if env_file:
        cli_cmd.append('-e')
        cli_cmd.append(env_file)

    try:
        subprocess.check_output(cli_cmd)
    except subprocess.CalledProcessError as cpe:
        print(cpe)
        raise cpe



from dagster import check

from dagster.core.types import Field, String
from dagster.core.types.field_utils import check_opt_field_param


class ResourceDefinition(object):
    def __init__(self, resource_fn, config_field=None, description=None):
        self.resource_fn = check.callable_param(resource_fn, 'resource_fn')
        self.config_field = check_opt_field_param(config_field, 'config_field')
        self.description = check.opt_str_param(description, 'description')

    @staticmethod
    def none_resource(description=None):
        return ResourceDefinition(resource_fn=lambda _init_context: None, description=description)

    @staticmethod
    def string_resource(description=None):
        return ResourceDefinition(
            resource_fn=lambda init_context: init_context.resource_config,
            config_field=Field(String),
            description=description,
        )


def resource(config_field=None, description=None):
    # This case is for when decorator is used bare, without arguments.
    # E.g. @resource versus @resource()
    if callable(config_field):
        return ResourceDefinition(resource_fn=config_field)

    def _wrap(resource_fn):
        return ResourceDefinition(resource_fn, config_field, description)

    return _wrap



import subprocess

from dagster import execute_pipeline
from dagster.tutorials.intro_tutorial.execution_context import (
    define_execution_context_pipeline_step_one,
    define_execution_context_pipeline_step_two,
    define_execution_context_pipeline_step_three,
)
from dagster.utils import script_relative_path


def test_execution_context():
    for pipeline_definition in [
        define_execution_context_pipeline_step_one,
        define_execution_context_pipeline_step_two,
        define_execution_context_pipeline_step_three,
    ]:
        pipeline = pipeline_definition()
        execute_pipeline(pipeline, {'context': {'default': {'config': {'log_level': 'DEBUG'}}}})


def test_execution_context_invocation_as_script():
    subprocess.check_output(
        [
            'python',
            script_relative_path('../../../dagster/tutorials/intro_tutorial/execution_context.py'),
        ]
    )



import pytest

from dagster import (
    DependencyDefinition,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    RunConfig,
    lambda_solid,
)

from dagster.core.errors import (
    DagsterExecutionStepNotFoundError,
    DagsterInvalidSubplanInputNotFoundError,
    DagsterInvalidSubplanMissingInputError,
    DagsterInvalidSubplanOutputNotFoundError,
)

from dagster.core.execute_marshalling import execute_marshalling, MarshalledOutput

from dagster.core.execution_plan.objects import StepKind
from dagster.core.types.runtime import resolve_to_runtime_type

from dagster.core.types.marshal import serialize_to_file, deserialize_from_file

from dagster.utils.test import get_temp_file_names


def define_inty_pipeline():
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(inputs=[InputDefinition('num', Int)], output=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    @lambda_solid
    def user_throw_exception():
        raise Exception('whoops')

    pipeline = PipelineDefinition(
        name='basic_external_plan_execution',
        solids=[return_one, add_one, user_throw_exception],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )
    return pipeline


def test_basic_pipeline_external_plan_execution():
    pipeline = define_inty_pipeline()

    with get_temp_file_names(2) as temp_files:

        temp_path, write_path = temp_files  # pylint: disable=W0632

        int_type = resolve_to_runtime_type(Int)

        serialize_to_file(int_type.serialization_strategy, 5, temp_path)

        step_events = execute_marshalling(
            pipeline,
            ['add_one.transform'],
            inputs_to_marshal={'add_one.transform': {'num': temp_path}},
            outputs_to_marshal={'add_one.transform': [MarshalledOutput('result', write_path)]},
        )

        assert deserialize_from_file(int_type.serialization_strategy, write_path) == 6

    assert len(step_events) == 1

    transform_step_output_event = step_events[0]
    assert transform_step_output_event.step_kind == StepKind.TRANSFORM
    assert transform_step_output_event.is_successful_output
    assert transform_step_output_event.step_output_data.output_name == 'result'
    assert transform_step_output_event.step_output_data.get_value() == 6


def test_external_execution_marshal_wrong_input_error():
    pipeline = define_inty_pipeline()

    with pytest.raises(DagsterInvalidSubplanInputNotFoundError) as exc_info:
        execute_marshalling(
            pipeline,
            ['add_one.transform'],
            inputs_to_marshal={'add_one.transform': {'nope': 'nope'}},
        )

    assert str(exc_info.value) == 'Input nope on add_one.transform does not exist.'
    assert exc_info.value.pipeline_name == pipeline.name
    assert exc_info.value.step_keys == ['add_one.transform']


def test_external_execution_step_for_input_missing():
    pipeline = define_inty_pipeline()

    with pytest.raises(DagsterExecutionStepNotFoundError) as exc_info:
        execute_marshalling(
            pipeline, ['add_one.transform'], inputs_to_marshal={'nope': {'nope': 'nope'}}
        )

    assert exc_info.value.step_key == 'nope'


def test_external_execution_input_marshal_code_error():
    pipeline = define_inty_pipeline()

    with pytest.raises(IOError):
        execute_marshalling(
            pipeline,
            ['add_one.transform'],
            inputs_to_marshal={'add_one.transform': {'num': 'nope'}},
        )

    # This throws outside of the normal execution process for now so this is expected
    with pytest.raises(IOError):
        execute_marshalling(
            pipeline,
            ['add_one.transform'],
            inputs_to_marshal={'add_one.transform': {'num': 'nope'}},
            run_config=RunConfig.nonthrowing_in_process(),
        )


def test_external_execution_step_for_output_missing():
    pipeline = define_inty_pipeline()

    with pytest.raises(DagsterExecutionStepNotFoundError):
        execute_marshalling(
            pipeline,
            ['add_one.transform'],
            inputs_to_marshal={'add_one.transform': {'num': 'nope'}},
            outputs_to_marshal={'nope': [MarshalledOutput('nope', 'nope')]},
        )


def test_external_execution_output_missing():
    pipeline = define_inty_pipeline()

    with pytest.raises(DagsterInvalidSubplanOutputNotFoundError):
        execute_marshalling(
            pipeline,
            ['add_one.transform'],
            outputs_to_marshal={'add_one.transform': [MarshalledOutput('nope', 'nope')]},
        )


def test_external_execution_marshal_output_code_error():
    pipeline = define_inty_pipeline()

    # guaranteed that folder does not exist
    hardcoded_uuid = '83fb4ace-5cab-459d-99b6-2ca9808c54a1'

    outputs_to_marshal = {
        'add_one.transform': [
            MarshalledOutput(
                output_name='result', marshalling_key='{uuid}/{uuid}'.format(uuid=hardcoded_uuid)
            )
        ]
    }

    with pytest.raises(IOError) as exc_info:
        execute_marshalling(
            pipeline,
            ['return_one.transform', 'add_one.transform'],
            outputs_to_marshal=outputs_to_marshal,
        )

    assert 'No such file or directory' in str(exc_info.value)

    # This throws outside of the normal execution process for now so this is expected
    with pytest.raises(IOError) as exc_info:
        execute_marshalling(
            pipeline,
            ['return_one.transform', 'add_one.transform'],
            outputs_to_marshal=outputs_to_marshal,
            run_config=RunConfig.nonthrowing_in_process(),
        )


def test_external_execution_output_code_error_throw_on_user_error():
    pipeline = define_inty_pipeline()

    with pytest.raises(Exception) as exc_info:
        execute_marshalling(pipeline, ['user_throw_exception.transform'])

    assert str(exc_info.value) == 'whoops'


def test_external_execution_output_code_error_no_throw_on_user_error():
    pipeline = define_inty_pipeline()

    step_events = execute_marshalling(
        pipeline, ['user_throw_exception.transform'], run_config=RunConfig.nonthrowing_in_process()
    )

    assert len(step_events) == 1
    step_event = step_events[0]
    assert step_event.step_failure_data
    assert step_event.step_failure_data.error_cls_name == 'Exception'
    assert step_event.step_failure_data.error_message == 'Exception: whoops\n'


def test_external_execution_unsatisfied_input_error():
    pipeline = define_inty_pipeline()

    with pytest.raises(DagsterInvalidSubplanMissingInputError) as exc_info:
        execute_marshalling(pipeline, ['add_one.transform'])

    assert exc_info.value.pipeline_name == 'basic_external_plan_execution'
    assert exc_info.value.step_keys == ['add_one.transform']
    assert exc_info.value.step.key == 'add_one.transform'
    assert exc_info.value.input_name == 'num'

    assert str(exc_info.value) == (
        "You have specified a subset execution on pipeline basic_external_plan_execution with "
        "step_keys ['add_one.transform']. You have failed to provide the required input num for "
        "step add_one.transform."
    )



from dagster import execute_pipeline
from dagster.tutorials.intro_tutorial.hello_world import define_hello_world_pipeline
from dagster.tutorials.utils import check_script, check_cli_execute_file_pipeline
from dagster.utils import script_relative_path


def test_tutorial_intro_tutorial_hello_world():
    pipeline = define_hello_world_pipeline()

    result = execute_pipeline(pipeline)

    assert result.success
    assert len(result.solid_result_list) == 1
    assert result.result_for_solid('hello_world').transformed_value() == 'hello'
    return result


def test_tutorial_intro_tutorial_hello_world_script():
    check_script(script_relative_path('../../../dagster/tutorials/intro_tutorial/hello_world.py'))


def test_tutorial_intro_tutorial_hello_world_cli():
    check_cli_execute_file_pipeline(
        script_relative_path('../../../dagster/tutorials/intro_tutorial/hello_world.py'),
        'define_hello_world_pipeline',
    )



from collections import namedtuple
from enum import Enum


from dagster import check
from dagster.core.definitions import Solid, PipelineDefinition
from dagster.core.execution_context import (
    SystemPipelineExecutionContext,
    SystemStepExecutionContext,
)
from dagster.core.types.runtime import RuntimeType
from dagster.core.utils import toposort
from dagster.utils import merge_dicts

from .intermediates_manager import StepOutputHandle


class StepOutputValue(namedtuple('_StepOutputValue', 'output_name value')):
    def __new__(cls, output_name, value):
        return super(StepOutputValue, cls).__new__(
            cls, output_name=check.str_param(output_name, 'output_name'), value=value
        )


class SingleOutputStepCreationData(namedtuple('SingleOutputStepCreationData', 'step output_name')):
    '''
    It is very common for step creation to involve processing a single value (e.g. an input thunk).
    This tuple is meant to be used by those functions to return both a new step and the output
    that deals with the value in question.
    '''

    @property
    def step_output_handle(self):
        return StepOutputHandle.from_step(self.step, self.output_name)


class StepOutputData:
    def __init__(self, step_output_handle, value_repr, intermediates_manager):
        self.step_output_handle = step_output_handle
        self._value_repr = value_repr
        self._intermediates_manager = intermediates_manager

    @property
    def output_name(self):
        return self.step_output_handle.output_name

    @property
    def value_repr(self):
        return self._value_repr

    def get_value(self):
        return self._intermediates_manager.get_value(self.step_output_handle)


class StepFailureData(namedtuple('_StepFailureData', 'error_message error_cls_name stack')):
    def __new__(cls, error_message, error_cls_name, stack):
        return super(StepFailureData, cls).__new__(
            cls,
            error_cls_name=check.str_param(error_cls_name, 'error_cls_name'),
            error_message=check.str_param(error_message, 'error_message'),
            stack=check.list_param(stack, 'stack', of_type=str),
        )


class ExecutionStepEventType(Enum):
    STEP_OUTPUT = 'STEP_OUTPUT'
    STEP_FAILURE = 'STEP_FAILURE'


class ExecutionStepEvent(
    namedtuple(
        '_ExecutionStepEvent',
        'event_type step_key solid_name step_kind step_output_data step_failure_data tags',
    )
):
    @staticmethod
    def from_step(event_type, step, step_output_data, step_failure_data, tags):
        return ExecutionStepEvent(
            event_type,
            step.key,
            step.solid.name,
            step.kind,
            step_output_data,
            step_failure_data,
            tags,
        )

    def __new__(
        cls, event_type, step_key, solid_name, step_kind, step_output_data, step_failure_data, tags
    ):
        return super(ExecutionStepEvent, cls).__new__(
            cls,
            check.inst_param(event_type, 'event_type', ExecutionStepEventType),
            check.str_param(step_key, 'step_key'),
            check.str_param(solid_name, 'solid_name'),
            check.inst_param(step_kind, 'step_kind', StepKind),
            check.opt_inst_param(step_output_data, 'step_output_data', StepOutputData),
            check.opt_inst_param(step_failure_data, 'step_failure_data', StepFailureData),
            check.dict_param(tags, 'tags'),
        )

    @property
    def is_step_success(self):
        return not self.is_step_failure

    @property
    def is_successful_output(self):
        return self.event_type == ExecutionStepEventType.STEP_OUTPUT

    @property
    def is_step_failure(self):
        return self.event_type == ExecutionStepEventType.STEP_FAILURE

    @staticmethod
    def step_output_event(step_context, step_output_data):
        check.inst_param(step_context, 'step_context', SystemStepExecutionContext)

        return ExecutionStepEvent.from_step(
            event_type=ExecutionStepEventType.STEP_OUTPUT,
            step=step_context.step,
            step_output_data=check.inst_param(step_output_data, 'step_output_data', StepOutputData),
            step_failure_data=None,
            tags=step_context.tags,
        )

    @staticmethod
    def step_failure_event(step_context, step_failure_data):
        check.inst_param(step_context, 'step_context', SystemStepExecutionContext)

        return ExecutionStepEvent.from_step(
            event_type=ExecutionStepEventType.STEP_FAILURE,
            step=step_context.step,
            step_output_data=None,
            step_failure_data=check.inst_param(
                step_failure_data, 'step_failure_data', StepFailureData
            ),
            tags=step_context.tags,
        )


class StepKind(Enum):
    TRANSFORM = 'TRANSFORM'
    INPUT_EXPECTATION = 'INPUT_EXPECTATION'
    OUTPUT_EXPECTATION = 'OUTPUT_EXPECTATION'
    JOIN = 'JOIN'
    SERIALIZE = 'SERIALIZE'
    INPUT_THUNK = 'INPUT_THUNK'
    MATERIALIZATION_THUNK = 'MATERIALIZATION_THUNK'
    VALUE_THUNK = 'VALUE_THUNK'
    UNMARSHAL_INPUT = 'UNMARSHAL_INPUT'
    MARSHAL_OUTPUT = 'MARSHAL_OUTPUT'


class StepInput(namedtuple('_StepInput', 'name runtime_type prev_output_handle')):
    def __new__(cls, name, runtime_type, prev_output_handle):
        return super(StepInput, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            runtime_type=check.inst_param(runtime_type, 'runtime_type', RuntimeType),
            prev_output_handle=check.inst_param(
                prev_output_handle, 'prev_output_handle', StepOutputHandle
            ),
        )


class StepOutput(namedtuple('_StepOutput', 'name runtime_type')):
    def __new__(cls, name, runtime_type):
        return super(StepOutput, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            runtime_type=check.inst_param(runtime_type, 'runtime_type', RuntimeType),
        )


class ExecutionStep(
    namedtuple(
        '_ExecutionStep',
        (
            'pipeline_context key step_inputs step_input_dict step_outputs step_output_dict '
            'compute_fn kind solid tags'
        ),
    )
):
    def __new__(
        cls, pipeline_context, key, step_inputs, step_outputs, compute_fn, kind, solid, tags=None
    ):
        return super(ExecutionStep, cls).__new__(
            cls,
            pipeline_context=check.inst_param(
                pipeline_context, 'pipeline_context', SystemPipelineExecutionContext
            ),
            key=check.str_param(key, 'key'),
            step_inputs=check.list_param(step_inputs, 'step_inputs', of_type=StepInput),
            step_input_dict={si.name: si for si in step_inputs},
            step_outputs=check.list_param(step_outputs, 'step_outputs', of_type=StepOutput),
            step_output_dict={so.name: so for so in step_outputs},
            compute_fn=check.callable_param(compute_fn, 'compute_fn'),
            kind=check.inst_param(kind, 'kind', StepKind),
            solid=check.inst_param(solid, 'solid', Solid),
            tags=merge_dicts(
                merge_dicts(
                    {
                        'step_key': key,
                        'pipeline': pipeline_context.pipeline_def.name,
                        'solid': solid.name,
                        'solid_definition': solid.definition.name,
                    },
                    check.opt_dict_param(tags, 'tags'),
                ),
                pipeline_context.tags,
            ),
        )

    @property
    def pipeline_name(self):
        return self.pipeline_context.pipeline_def.name

    @property
    def solid_name(self):
        return self.solid.name

    @property
    def solid_definition_name(self):
        return self.solid.definition.name

    def __getnewargs__(self):
        return (
            self.key,
            self.step_inputs,
            self.step_outputs,
            self.compute_fn,
            self.kind,
            self.solid,
        )

    def with_new_inputs(self, step_inputs):
        return ExecutionStep(
            pipeline_context=self.pipeline_context,
            key=self.key,
            step_inputs=step_inputs,
            step_outputs=self.step_outputs,
            compute_fn=self.compute_fn,
            kind=self.kind,
            solid=self.solid,
            tags=self.tags,
        )

    def has_step_output(self, name):
        check.str_param(name, 'name')
        return name in self.step_output_dict

    def step_output_named(self, name):
        check.str_param(name, 'name')
        return self.step_output_dict[name]

    def has_step_input(self, name):
        check.str_param(name, 'name')
        return name in self.step_input_dict

    def step_input_named(self, name):
        check.str_param(name, 'name')
        return self.step_input_dict[name]


class ExecutionValueSubplan(
    namedtuple('ExecutionValueSubplan', 'steps terminal_step_output_handle')
):
    '''
    A frequent pattern in the execution engine is to take a single value
    (e.g. an input or an output of a transform) and then flow that value
    value through a sequence of system-injected steps (e.g. expectations
    or materializations). This object captures that pattern. It contains
    all of the steps that comprise that Subplan and also a single output
    handle that points to output that further steps down the plan can
    depend on.
    '''

    def __new__(cls, steps, terminal_step_output_handle):
        return super(ExecutionValueSubplan, cls).__new__(
            cls,
            check.list_param(steps, 'steps', of_type=ExecutionStep),
            check.inst_param(
                terminal_step_output_handle, 'terminal_step_output_handle', StepOutputHandle
            ),
        )

    @staticmethod
    def empty(terminal_step_output_handle):
        return ExecutionValueSubplan([], terminal_step_output_handle)


class ExecutionPlan(namedtuple('_ExecutionPlan', 'pipeline_def step_dict, deps, steps')):
    def __new__(cls, pipeline_def, step_dict, deps):
        return super(ExecutionPlan, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            step_dict=check.dict_param(
                step_dict, 'step_dict', key_type=str, value_type=ExecutionStep
            ),
            deps=check.dict_param(deps, 'deps', key_type=str, value_type=set),
            steps=list(step_dict.values()),
        )

    def get_step_output(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        step = self.get_step_by_key(step_output_handle.step_key)
        return step.step_output_named(step_output_handle.output_name)

    def has_step(self, key):
        check.str_param(key, 'key')
        return key in self.step_dict

    def get_step_by_key(self, key):
        check.str_param(key, 'key')
        return self.step_dict[key]

    def topological_steps(self):
        return [step for step_level in self.topological_step_levels() for step in step_level]

    def topological_step_levels(self):
        return [
            [self.step_dict[step_key] for step_key in step_key_level]
            for step_key_level in toposort(self.deps)
        ]



from dagster import execute_pipeline
from dagster.tutorials.intro_tutorial.hello_dag import define_hello_dag_pipeline
from dagster.tutorials.utils import check_cli_execute_file_pipeline
from dagster.utils import script_relative_path


def test_intro_tutorial_hello_dag():
    pipeline = define_hello_dag_pipeline()

    result = execute_pipeline(pipeline)

    assert result.success
    assert len(result.solid_result_list) == 2
    assert result.result_for_solid('solid_one').transformed_value() == 'foo'
    assert result.result_for_solid('solid_two').transformed_value() == 'foofoo'
    return result


def test_tutorial_cli_hello_dag():
    check_cli_execute_file_pipeline(
        script_relative_path('../../../dagster/tutorials/intro_tutorial/hello_dag.py'),
        'define_hello_dag_pipeline',
    )



from collections import defaultdict, namedtuple

from dagster import check

from .solid import SolidDefinition

from .input import InputDefinition

from .output import OutputDefinition

from .utils import DEFAULT_OUTPUT, struct_to_string


class SolidInstance(namedtuple('Solid', 'name alias')):
    '''
    A solid identifier in a dependency structure. Allows supplying parameters to the solid,
    like the alias.

    Example:

        .. code-block:: python

            pipeline = Pipeline(
                solids=[solid_1, solid_2]
                dependencies={
                    SolidInstance('solid_2', alias='other_name') : {
                        'input_name' : DependencyDefinition('solid_2'),
                    },
                    'solid_1' : {
                        'input_name': DependencyDefinition('other_name'),
                    },
                }
            )
    '''

    def __new__(cls, name, alias=None):
        name = check.str_param(name, 'name')
        alias = check.opt_str_param(alias, 'alias')
        return super(cls, SolidInstance).__new__(cls, name, alias)


class Solid(object):
    '''
    Solid instance within a pipeline. Defined by it's name inside the pipeline.

    Attributes:
        name (str):
            Name of the solid inside the pipeline. Must be unique per-pipeline.
        definition (SolidDefinition):
            Definition of the solid.
    '''

    def __init__(self, name, definition):
        self.name = check.str_param(name, 'name')
        self.definition = check.inst_param(definition, 'definition', SolidDefinition)

        input_handles = {}
        for input_def in self.definition.input_defs:
            input_handles[input_def.name] = SolidInputHandle(self, input_def)

        self._input_handles = input_handles

        output_handles = {}
        for output_def in self.definition.output_defs:
            output_handles[output_def.name] = SolidOutputHandle(self, output_def)

        self._output_handles = output_handles

    def input_handles(self):
        return self._input_handles.values()

    def output_handles(self):
        return self._output_handles.values()

    def input_handle(self, name):
        check.str_param(name, 'name')
        return self._input_handles[name]

    def output_handle(self, name):
        check.str_param(name, 'name')
        return self._output_handles[name]

    def has_input(self, name):
        return self.definition.has_input(name)

    def input_def_named(self, name):
        return self.definition.input_def_named(name)

    def has_output(self, name):
        return self.definition.has_output(name)

    def output_def_named(self, name):
        return self.definition.output_def_named(name)

    @property
    def input_defs(self):
        return self.definition.input_defs

    @property
    def output_defs(self):
        return self.definition.output_defs


class SolidInputHandle(namedtuple('_SolidInputHandle', 'solid input_def')):
    def __new__(cls, solid, input_def):
        return super(SolidInputHandle, cls).__new__(
            cls,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(input_def, 'input_def', InputDefinition),
        )

    def _inner_str(self):
        return struct_to_string(
            'SolidInputHandle',
            solid_name=self.solid.name,
            definition_name=self.solid.definition.name,
            input_name=self.input_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.solid.name, self.input_def.name))

    def __eq__(self, other):
        return self.solid.name == other.solid.name and self.input_def.name == other.input_def.name


class SolidOutputHandle(namedtuple('_SolidOutputHandle', 'solid output_def')):
    def __new__(cls, solid, output_def):
        return super(SolidOutputHandle, cls).__new__(
            cls,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(output_def, 'output_def', OutputDefinition),
        )

    def _inner_str(self):
        return struct_to_string(
            'SolidOutputHandle',
            solid_name=self.solid.name,
            definition_name=self.solid.definition.name,
            output_name=self.output_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.solid.name, self.output_def.name))

    def __eq__(self, other):
        return self.solid.name == other.solid.name and self.output_def.name == other.output_def.name


class InputToOutputHandleDict(dict):
    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidInputHandle)
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidInputHandle)
        check.inst_param(val, 'val', SolidOutputHandle)
        return dict.__setitem__(self, key, val)


def _create_handle_dict(solid_dict, dep_dict):
    check.dict_param(solid_dict, 'solid_dict', key_type=str, value_type=Solid)
    check.two_dim_dict_param(dep_dict, 'dep_dict', value_type=DependencyDefinition)

    handle_dict = InputToOutputHandleDict()

    for solid_name, input_dict in dep_dict.items():
        for input_name, dep_def in input_dict.items():
            from_solid = solid_dict[solid_name]
            to_solid = solid_dict[dep_def.solid]
            handle_dict[from_solid.input_handle(input_name)] = to_solid.output_handle(
                dep_def.output
            )

    return handle_dict


class DependencyStructure(object):
    @staticmethod
    def from_definitions(solids, dep_dict):
        return DependencyStructure(_create_handle_dict(solids, dep_dict))

    def __init__(self, handle_dict):
        self._handle_dict = check.inst_param(handle_dict, 'handle_dict', InputToOutputHandleDict)

    def has_dep(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return solid_input_handle in self._handle_dict

    def deps_of_solid(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        return list(handles[1] for handles in self.__gen_deps_of_solid(solid_name))

    def deps_of_solid_with_input(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        return dict(self.__gen_deps_of_solid(solid_name))

    def __gen_deps_of_solid(self, solid_name):
        for input_handle, output_handle in self._handle_dict.items():
            if input_handle.solid.name == solid_name:
                yield (input_handle, output_handle)

    def depended_by_of_solid(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        result = defaultdict(list)
        for input_handle, output_handle in self._handle_dict.items():
            if output_handle.solid.name == solid_name:
                result[output_handle].append(input_handle)
        return result

    def get_dep(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return self._handle_dict[solid_input_handle]

    def input_handles(self):
        return list(self._handle_dict.keys())

    def items(self):
        return self._handle_dict.items()


class DependencyDefinition(namedtuple('_DependencyDefinition', 'solid output description')):
    '''Dependency definitions represent an edge in the DAG of solids. This object is
    used with a dictionary structure (whose keys represent solid/input where the dependency
    comes from) so this object only contains the target dependency information.

    Attributes:
        solid (str):
            The name of the solid that is the target of the dependency.
            This is the solid where the value passed between the solids
            comes from.
        output (str):
            The name of the output that is the target of the dependency.
            Defaults to "result", the default output name of solids with a single output.
        description (str):
            Description of this dependency. Optional.
    '''

    def __new__(cls, solid, output=DEFAULT_OUTPUT, description=None):
        return super(DependencyDefinition, cls).__new__(
            cls,
            check.str_param(solid, 'solid'),
            check.str_param(output, 'output'),
            check.opt_str_param(description, 'description'),
        )



from dagster import check

from dagster.core.execution_context import (
    SystemStepExecutionContext,
    SystemPipelineExecutionContext,
)

from dagster.core.user_context import ExpectationExecutionContext

from dagster.core.definitions import ExpectationDefinition, InputDefinition, OutputDefinition, Solid

from dagster.core.errors import DagsterExpectationFailedError

from .objects import (
    ExecutionStep,
    ExecutionValueSubplan,
    StepInput,
    StepOutput,
    StepOutputHandle,
    StepOutputValue,
    StepKind,
)

from .utility import create_joining_subplan

EXPECTATION_INPUT = 'expectation_input'
EXPECTATION_VALUE_OUTPUT = 'expectation_value'


def _create_expectation_lambda(solid, inout_def, expectation_def, internal_output_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(expectation_def, 'expectations_def', ExpectationDefinition)
    check.str_param(internal_output_name, 'internal_output_name')

    def _do_expectation(expectation_context, inputs):
        check.inst_param(expectation_context, 'step_context', SystemStepExecutionContext)
        value = inputs[EXPECTATION_INPUT]
        expectation_context = expectation_context.for_expectation(inout_def, expectation_def)
        expt_result = expectation_def.expectation_fn(
            ExpectationExecutionContext(expectation_context), value
        )
        # step_context,  value)
        if expt_result.success:
            expectation_context.log.debug(
                'Expectation {key} succeeded on {value}.'.format(
                    key=expectation_context.step.key, value=value
                )
            )
            yield StepOutputValue(output_name=internal_output_name, value=inputs[EXPECTATION_INPUT])
        else:
            expectation_context.log.debug(
                'Expectation {key} failed on {value}.'.format(
                    key=expectation_context.step.key, value=value
                )
            )
            raise DagsterExpectationFailedError(expectation_context, value)

    return _do_expectation


def create_expectations_subplan(pipeline_context, solid, inout_def, prev_step_output_handle, kind):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(kind, 'kind', StepKind)

    input_expect_steps = []
    for expectation_def in inout_def.expectations:
        expect_step = create_expectation_step(
            pipeline_context=pipeline_context,
            solid=solid,
            expectation_def=expectation_def,
            key='{solid}.{desc_key}.{inout_name}.expectation.{expectation_name}'.format(
                solid=solid.name,
                desc_key=inout_def.descriptive_key,
                inout_name=inout_def.name,
                expectation_name=expectation_def.name,
            ),
            kind=kind,
            prev_step_output_handle=prev_step_output_handle,
            inout_def=inout_def,
        )
        input_expect_steps.append(expect_step)

    return create_joining_subplan(
        pipeline_context,
        solid,
        '{solid}.{desc_key}.{inout_name}.expectations.join'.format(
            solid=solid.name, desc_key=inout_def.descriptive_key, inout_name=inout_def.name
        ),
        input_expect_steps,
        EXPECTATION_VALUE_OUTPUT,
    )


def create_expectation_step(
    pipeline_context, solid, expectation_def, key, kind, prev_step_output_handle, inout_def
):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(expectation_def, 'input_expct_def', ExpectationDefinition)
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))

    value_type = inout_def.runtime_type

    return ExecutionStep(
        pipeline_context=pipeline_context,
        key=key,
        step_inputs=[
            StepInput(
                name=EXPECTATION_INPUT,
                runtime_type=value_type,
                prev_output_handle=prev_step_output_handle,
            )
        ],
        step_outputs=[StepOutput(name=EXPECTATION_VALUE_OUTPUT, runtime_type=value_type)],
        compute_fn=_create_expectation_lambda(
            solid, inout_def, expectation_def, EXPECTATION_VALUE_OUTPUT
        ),
        kind=kind,
        solid=solid,
        tags={'expectation': expectation_def.name, inout_def.descriptive_key: inout_def.name},
    )


def decorate_with_expectations(pipeline_context, solid, transform_step, output_def):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(transform_step, 'transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    if pipeline_context.environment_config.expectations.evaluate and output_def.expectations:
        return create_expectations_subplan(
            pipeline_context,
            solid,
            output_def,
            StepOutputHandle.from_step(transform_step, output_def.name),
            kind=StepKind.OUTPUT_EXPECTATION,
        )
    else:
        return ExecutionValueSubplan.empty(
            StepOutputHandle.from_step(transform_step, output_def.name)
        )



import os

from dagster import check

from dagster.core.execution_context import (
    MultiprocessExecutorConfig,
    SystemPipelineExecutionContext,
)


from .create import create_execution_plan_core
from .intermediates_manager import FileSystemIntermediateManager
from .objects import ExecutionPlan, ExecutionStepEvent
from .simple_engine import start_inprocess_executor

from .child_process_executor import ChildProcessCommand, execute_child_process_command


class InProcessExecutorChildProcessCommand(ChildProcessCommand):
    def __init__(self, environment_dict, run_config, step_key):
        self.environment_dict = environment_dict
        self.run_config = run_config
        self.step_key = step_key

    def execute(self):
        from dagster.core.execution import yield_pipeline_execution_context

        check.inst(self.run_config.executor_config, MultiprocessExecutorConfig)
        pipeline = self.run_config.executor_config.pipeline_fn()

        with yield_pipeline_execution_context(
            pipeline, self.environment_dict, self.run_config.with_tags(pid=str(os.getpid()))
        ) as pipeline_context:

            intermediates_manager = FileSystemIntermediateManager(pipeline_context.files)

            execution_plan = create_execution_plan_core(pipeline_context)

            for step_event in start_inprocess_executor(
                pipeline_context,
                execution_plan,
                intermediates_manager,
                step_keys_to_execute=[self.step_key],
            ):
                yield step_event


def execute_step_out_of_process(step_context, step):
    check.invariant(
        not step_context.run_config.loggers,
        'Cannot inject loggers via RunConfig with the Multiprocess executor',
    )

    command = InProcessExecutorChildProcessCommand(
        step_context.environment_dict, step_context.run_config, step.key
    )

    for step_event in execute_child_process_command(command):
        yield step_event


def _create_input_values(step_input_meta_dict, manager):
    input_values = {}
    for step_input_name, prev_output_handle_meta in step_input_meta_dict.items():
        input_value = manager.get_value(prev_output_handle_meta)
        input_values[step_input_name] = input_value
    return input_values


def multiprocess_execute_plan(pipeline_context, execution_plan):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)

    step_levels = execution_plan.topological_step_levels()

    intermediates_manager = FileSystemIntermediateManager(pipeline_context.files)

    # It would be good to implement a reference tracking algorithm here so we could
    # garbage collection results that are no longer needed by any steps
    # https://github.com/dagster-io/dagster/issues/811

    for step_level in step_levels:
        for step in step_level:
            step_context = pipeline_context.for_step(step)

            if not intermediates_manager.all_inputs_covered(step):
                expected_outputs = [ni.prev_output_handle for ni in step.step_inputs]

                step_context.log.debug(
                    (
                        'Not all inputs covered for {step}. Not executing.'
                        'Outputs need for inputs {expected_outputs}'
                    ).format(expected_outputs=expected_outputs, step=step.key)
                )
                continue

            for step_event in check.generator(execute_step_out_of_process(step_context, step)):
                check.inst(step_event, ExecutionStepEvent)
                yield step_event



'''
This module contains the execution context objects that are internal to the system.
Not every property on these should be exposed to random Jane or Joe dagster user
so we have a different layer of objects that encode the explicit public API
in the user_context module
'''
from collections import namedtuple
import uuid

from dagster import check
from dagster.utils import merge_dicts

from .definitions.expectation import ExpectationDefinition
from .definitions.input import InputDefinition
from .definitions.output import OutputDefinition
from .events import ExecutionEvents
from .files import FileStore
from .log import DagsterLog
from .system_config.objects import EnvironmentConfig
from .types.marshal import PersistenceStrategy


class ExecutorConfig:
    pass


class InProcessExecutorConfig(ExecutorConfig):
    def __init__(self, throw_on_user_error=True, inmem_intermediates_manager=None):
        from .execution_plan.intermediates_manager import InMemoryIntermediatesManager

        self.throw_on_user_error = check.bool_param(throw_on_user_error, 'throw_on_user_error')
        self.inmem_intermediates_manager = check.opt_inst_param(
            inmem_intermediates_manager,
            'inmem_intermediates_manager',
            InMemoryIntermediatesManager,
            InMemoryIntermediatesManager(),
        )


class MultiprocessExecutorConfig(ExecutorConfig):
    def __init__(self, pipeline_fn):
        self.pipeline_fn = check.callable_param(pipeline_fn, 'pipeline_fn')
        self.throw_on_user_error = False


def make_new_run_id():
    return str(uuid.uuid4())


class RunConfig(namedtuple('_RunConfig', 'run_id tags event_callback loggers executor_config')):
    def __new__(
        cls, run_id=None, tags=None, event_callback=None, loggers=None, executor_config=None
    ):
        return super(RunConfig, cls).__new__(
            cls,
            run_id=check.str_param(run_id, 'run_id') if run_id else make_new_run_id(),
            tags=check.opt_dict_param(tags, 'tags', key_type=str, value_type=str),
            event_callback=check.opt_callable_param(event_callback, 'event_callback'),
            loggers=check.opt_list_param(loggers, 'loggers'),
            executor_config=check.opt_inst_param(
                executor_config, 'executor_config', ExecutorConfig, InProcessExecutorConfig()
            ),
        )

    @staticmethod
    def nonthrowing_in_process():
        return RunConfig(executor_config=InProcessExecutorConfig(throw_on_user_error=False))

    def with_tags(self, **tags):
        return RunConfig(
            run_id=self.run_id,
            event_callback=self.event_callback,
            loggers=self.loggers,
            tags=merge_dicts(self.tags, tags),
            executor_config=self.executor_config,
        )


class SystemPipelineExecutionContextData(
    namedtuple(
        '_SystemPipelineExecutionContextData',
        'run_config resources environment_config persistence_strategy pipeline_def files',
    )
):
    '''
    PipelineContextData is the data that remains context throughtout the entire execution
    of a pipeline.
    '''

    def __new__(
        cls, run_config, resources, environment_config, persistence_strategy, pipeline_def, files
    ):
        from .definitions.pipeline import PipelineDefinition

        return super(SystemPipelineExecutionContextData, cls).__new__(
            cls,
            run_config=check.inst_param(run_config, 'run_config', RunConfig),
            resources=resources,
            environment_config=check.inst_param(
                environment_config, 'environment_config', EnvironmentConfig
            ),
            persistence_strategy=check.inst_param(
                persistence_strategy, 'persistence_strategy', PersistenceStrategy
            ),
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            files=check.inst_param(files, 'files', FileStore),
        )

    @property
    def run_id(self):
        return self.run_config.run_id

    @property
    def event_callback(self):
        return self.run_config.event_callback

    @property
    def environment_dict(self):
        return self.environment_config.original_config_dict


class SystemPipelineExecutionContext(object):
    __slots__ = ['_pipeline_context_data', '_tags', '_log', '_legacy_context', '_events']

    def __init__(self, pipeline_context_data, tags, log):
        self._pipeline_context_data = check.inst_param(
            pipeline_context_data, 'pipeline_context_data', SystemPipelineExecutionContextData
        )
        self._tags = check.dict_param(tags, 'tags')
        self._log = check.inst_param(log, 'log', DagsterLog)
        self._events = ExecutionEvents(pipeline_context_data.pipeline_def.name, self._log)

    def for_step(self, step):
        from .execution_plan.objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)

        tags = merge_dicts(self.tags, step.tags)
        log = DagsterLog(self.run_id, tags, self.log.loggers)
        return SystemStepExecutionContext(self._pipeline_context_data, tags, log, step)

    @property
    def executor_config(self):
        return self.run_config.executor_config

    @property
    def run_config(self):
        return self._pipeline_context_data.run_config

    @property
    def resources(self):
        return self._pipeline_context_data.resources

    @property
    def run_id(self):
        return self._pipeline_context_data.run_id

    @property
    def environment_dict(self):
        return self._pipeline_context_data.environment_dict

    @property
    def environment_config(self):
        return self._pipeline_context_data.environment_config

    @property
    def tags(self):
        return self._tags

    def has_tag(self, key):
        check.str_param(key, 'key')
        return key in self._tags

    def get_tag(self, key):
        check.str_param(key, 'key')
        return self._tags[key]

    @property
    def persistence_strategy(self):
        return self._pipeline_context_data.persistence_strategy

    @property
    def pipeline_def(self):
        return self._pipeline_context_data.pipeline_def

    @property
    def events(self):
        return self._events

    @property
    def event_callback(self):
        return self._pipeline_context_data.event_callback

    def has_event_callback(self):
        return self._pipeline_context_data.event_callback is not None

    @property
    def log(self):
        return self._log

    @property
    def files(self):
        return self._pipeline_context_data.files


class SystemStepExecutionContext(SystemPipelineExecutionContext):
    __slots__ = ['_step']

    def __init__(self, pipeline_context_data, tags, log, step):
        from .execution_plan.objects import ExecutionStep

        self._step = check.inst_param(step, 'step', ExecutionStep)
        super(SystemStepExecutionContext, self).__init__(pipeline_context_data, tags, log)

    def for_transform(self):
        return SystemTransformExecutionContext(
            self._pipeline_context_data, self.tags, self.log, self.step
        )

    def for_expectation(self, inout_def, expectation_def):
        return SystemExpectationExecutionContext(
            self._pipeline_context_data, self.tags, self.log, self.step, inout_def, expectation_def
        )

    @property
    def step(self):
        return self._step

    @property
    def solid_def(self):
        return self._step.solid.definition

    @property
    def solid(self):
        return self._step.solid


class SystemTransformExecutionContext(SystemStepExecutionContext):
    @property
    def solid_config(self):
        solid_config = self.environment_config.solids.get(self.solid.name)
        return solid_config.config if solid_config else None


class SystemExpectationExecutionContext(SystemStepExecutionContext):
    __slots__ = ['_inout_def', '_expectation_def']

    def __init__(self, pipeline_context_data, tags, log, step, inout_def, expectation_def):
        self._expectation_def = check.inst_param(
            expectation_def, 'expectation_def', ExpectationDefinition
        )
        self._inout_def = check.inst_param(
            inout_def, 'inout_def', (InputDefinition, OutputDefinition)
        )
        super(SystemExpectationExecutionContext, self).__init__(
            pipeline_context_data, tags, log, step
        )

    @property
    def expectation_def(self):
        return self._expectation_def

    @property
    def inout_def(self):
        return self._inout_def



from dagster import check

from dagster.core.definitions import Solid, OutputDefinition
from dagster.core.execution_context import SystemPipelineExecutionContext

from dagster.core.types.runtime import RuntimeType

from .objects import (
    ExecutionStep,
    ExecutionValueSubplan,
    StepInput,
    StepOutput,
    StepOutputValue,
    StepKind,
)

from .utility import create_joining_subplan

MATERIALIZATION_THUNK_INPUT = 'materialization_thunk_input'
MATERIALIZATION_THUNK_OUTPUT = 'materialization_thunk_output'


def _create_materialization_lambda(runtime_type, config_spec):
    check.inst_param(runtime_type, 'runtime_type', RuntimeType)
    check.invariant(runtime_type.output_schema, 'Must have output schema')

    def _fn(_step_context, inputs):
        runtime_value = inputs[MATERIALIZATION_THUNK_INPUT]
        runtime_type.output_schema.materialize_runtime_value(config_spec, runtime_value)
        yield StepOutputValue(output_name=MATERIALIZATION_THUNK_OUTPUT, value=runtime_value)

    return _fn


def configs_for_output(solid, solid_config, output_def):
    for output_spec in solid_config.outputs:
        check.invariant(len(output_spec) == 1)
        output_name, output_spec = list(output_spec.items())[0]
        check.invariant(solid.has_output(output_name))
        if output_name == output_def.name:
            yield output_spec


def decorate_with_output_materializations(pipeline_context, solid, output_def, subplan):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(output_def, 'output_def', OutputDefinition)
    check.inst_param(subplan, 'subplan', ExecutionValueSubplan)

    solid_config = pipeline_context.environment_config.solids.get(solid.name)

    if not (solid_config and solid_config.outputs):
        return subplan

    new_steps = []

    for mat_count, output_spec in enumerate(configs_for_output(solid, solid_config, output_def)):
        new_steps.append(
            ExecutionStep(
                pipeline_context=pipeline_context,
                key='{solid}.materialization.output.{output}.{mat_count}'.format(
                    solid=solid.name, output=output_def.name, mat_count=mat_count
                ),
                step_inputs=[
                    StepInput(
                        name=MATERIALIZATION_THUNK_INPUT,
                        runtime_type=output_def.runtime_type,
                        prev_output_handle=subplan.terminal_step_output_handle,
                    )
                ],
                step_outputs=[
                    StepOutput(
                        name=MATERIALIZATION_THUNK_OUTPUT, runtime_type=output_def.runtime_type
                    )
                ],
                kind=StepKind.MATERIALIZATION_THUNK,
                solid=solid,
                compute_fn=_create_materialization_lambda(output_def.runtime_type, output_spec),
                # tags=plan_builder.get_tags(),
            )
        )

    return create_joining_subplan(
        pipeline_context,
        solid,
        '{solid}.materialization.output.{output}.join'.format(
            solid=solid.name, output=output_def.name
        ),
        new_steps,
        MATERIALIZATION_THUNK_OUTPUT,
    )



from collections import namedtuple
from functools import wraps
import inspect

from dagster import check

from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError

from dagster.core.types.field_utils import check_opt_field_param

from . import InputDefinition, OutputDefinition, Result, SolidDefinition

if hasattr(inspect, 'signature'):
    funcsigs = inspect
else:
    import funcsigs

# Error messages are long
# pylint: disable=C0301


class MultipleResults(namedtuple('_MultipleResults', 'results')):
    '''A shortcut to output multiple results.

    When using the :py:func:`@solid <dagster.solid>` API, you may return an instance of
    ``MultipleResults`` from a decorated transform function instead of yielding multiple results.

    Attributes:
      results (list[Result]): list of :py:class:`Result`

    Examples:

        .. code-block:: python

            @solid(outputs=[
                OutputDefinition(name='foo'),
                OutputDefinition(name='bar'),
            ])
            def my_solid():
                return MultipleResults(
                    Result('Barb', 'foo'),
                    Result('Glarb', 'bar'),
                )


            @solid(outputs=[
                OutputDefinition(name='foo'),
                OutputDefinition(name='bar'),
            ])
            def my_solid_from_dict():
                return MultipleResults.from_dict({
                'foo': 'Barb',
                'bar': 'Glarb',
                })

    '''

    def __new__(cls, *results):
        return super(MultipleResults, cls).__new__(
            cls,
            # XXX(freiksenet): should check.list_param accept tuples from rest?
            check.opt_list_param(list(results), 'results', Result),
        )

    @staticmethod
    def from_dict(result_dict):
        '''Create a new ``MultipleResults`` object from a dictionary.
        
        Keys of the dictionary are unpacked into result names.
        
        Args:
            result_dict (dict) - The dictionary to unpack.
        
        Returns:
            (:py:class:`MultipleResults <dagster.MultipleResults>`) A new ``MultipleResults`` object

        '''
        check.dict_param(result_dict, 'result_dict', key_type=str)
        results = []
        for name, value in result_dict.items():
            results.append(Result(value, name))
        return MultipleResults(*results)


class _LambdaSolid(object):
    def __init__(self, name=None, inputs=None, output=None, description=None):
        self.name = check.opt_str_param(name, 'name')
        self.input_defs = check.opt_list_param(inputs, 'inputs', InputDefinition)
        self.output_def = check.inst_param(output, 'output', OutputDefinition)
        self.description = check.opt_str_param(description, 'description')

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        _validate_transform_fn(self.name, fn, self.input_defs)
        transform_fn = _create_lambda_solid_transform_wrapper(fn, self.input_defs, self.output_def)
        return SolidDefinition(
            name=self.name,
            inputs=self.input_defs,
            outputs=[self.output_def],
            transform_fn=transform_fn,
            description=self.description,
        )


class _Solid(object):
    def __init__(self, name=None, inputs=None, outputs=None, description=None, config_field=None):
        self.name = check.opt_str_param(name, 'name')
        self.input_defs = check.opt_list_param(inputs, 'inputs', InputDefinition)
        outputs = outputs or ([OutputDefinition()] if outputs is None else [])
        self.outputs = check.list_param(outputs, 'outputs', OutputDefinition)
        self.description = check.opt_str_param(description, 'description')
        self.config_field = check_opt_field_param(config_field, 'config_field')

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        _validate_transform_fn(self.name, fn, self.input_defs, [('info', 'context')])
        transform_fn = _create_solid_transform_wrapper(fn, self.input_defs, self.outputs)
        return SolidDefinition(
            name=self.name,
            inputs=self.input_defs,
            outputs=self.outputs,
            transform_fn=transform_fn,
            config_field=self.config_field,
            description=self.description,
        )


def lambda_solid(name=None, inputs=None, output=None, description=None):
    '''(decorator) Create a simple solid.

    This shortcut allows the creation of simple solids that do not require
    configuration and whose implementations do not require a context.

    Lambda solids take inputs and produce a single output. The body of the function
    should return a single value.

    Args:
        name (str): Name of solid.
        inputs (list[InputDefinition]): List of inputs.
        output (OutputDefinition): The output of the solid. Defaults to ``OutputDefinition()``.
        description (str): Solid description.

    Examples:

        .. code-block:: python

            @lambda_solid
            def hello_world():
                return 'hello'

            @lambda_solid(inputs=[InputDefinition(name='foo')])
            def hello_world(foo):
                return foo

    '''
    output = output or OutputDefinition()

    if callable(name):
        check.invariant(inputs is None)
        check.invariant(description is None)
        return _LambdaSolid(output=output)(name)

    return _LambdaSolid(name=name, inputs=inputs, output=output, description=description)


def solid(name=None, inputs=None, outputs=None, config_field=None, description=None):
    '''(decorator) Create a solid with specified parameters.

    This shortcut simplifies the core solid API by exploding arguments into kwargs of the
    transform function and omitting additional parameters when they are not needed.
    Parameters are otherwise as in the core API, :py:class:`SolidDefinition`.

    The decorated function will be used as the solid's transform function. Unlike in the core API,
    the transform function does not have to yield :py:class:`Result` object directly. Several
    simpler alternatives are available:

    1. Return a value. This is returned as a :py:class:`Result` for a single output solid.
    2. Return a :py:class:`Result`. Works like yielding result.
    3. Return an instance of :py:class:`MultipleResults`. Works like yielding several results for
       multiple outputs. Useful for solids that have multiple outputs.
    4. Yield :py:class:`Result`. Same as default transform behaviour.

    Args:
        name (str): Name of solid.
        inputs (list[InputDefinition]): List of inputs.
        outputs (list[OutputDefinition]): List of outputs.
        config_field (Field):
            The configuration for this solid.
        description (str): Description of this solid.

    Examples:

        .. code-block:: python

            @solid
            def hello_world(_context):
                print('hello')

            @solid()
            def hello_world(_context):
                print('hello')

            @solid(outputs=[OutputDefinition()])
            def hello_world(_context):
                return {'foo': 'bar'}

            @solid(outputs=[OutputDefinition()])
            def hello_world(_context):
                return Result(value={'foo': 'bar'})

            @solid(outputs=[OutputDefinition()])
            def hello_world(_context):
                yield Result(value={'foo': 'bar'})

            @solid(outputs=[
                OutputDefinition(name="left"),
                OutputDefinition(name="right"),
            ])
            def hello_world(_context):
                return MultipleResults.from_dict({
                    'left': {'foo': 'left'},
                    'right': {'foo': 'right'},
                })

            @solid(
                inputs=[InputDefinition(name="foo")],
                outputs=[OutputDefinition()]
            )
            def hello_world(_context, foo):
                return foo

            @solid(
                inputs=[InputDefinition(name="foo")],
                outputs=[OutputDefinition()],
            )
            def hello_world(context, foo):
                context.log.info('log something')
                return foo

            @solid(
                inputs=[InputDefinition(name="foo")],
                outputs=[OutputDefinition()],
                config_field=Field(types.Dict({'str_value' : Field(types.String)})),
            )
            def hello_world(context, foo):
                # context.solid_config is a dictionary with 'str_value' key
                return foo + context.solid_config['str_value']

    '''
    # This case is for when decorator is used bare, without arguments. e.g. @solid versus @solid()
    if callable(name):
        check.invariant(inputs is None)
        check.invariant(outputs is None)
        check.invariant(description is None)
        check.invariant(config_field is None)
        return _Solid()(name)

    return _Solid(
        name=name,
        inputs=inputs,
        outputs=outputs,
        config_field=config_field,
        description=description,
    )


def _create_lambda_solid_transform_wrapper(fn, input_defs, output_def):
    check.callable_param(fn, 'fn')
    check.list_param(input_defs, 'input_defs', of_type=InputDefinition)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    input_names = [input_def.name for input_def in input_defs]

    @wraps(fn)
    def transform(_context, inputs):
        kwargs = {}
        for input_name in input_names:
            kwargs[input_name] = inputs[input_name]

        result = fn(**kwargs)
        yield Result(value=result, output_name=output_def.name)

    return transform


def _create_solid_transform_wrapper(fn, input_defs, output_defs):
    check.callable_param(fn, 'fn')
    check.list_param(input_defs, 'input_defs', of_type=InputDefinition)
    check.list_param(output_defs, 'output_defs', of_type=OutputDefinition)

    input_names = [input_def.name for input_def in input_defs]

    @wraps(fn)
    def transform(context, inputs):
        kwargs = {}
        for input_name in input_names:
            kwargs[input_name] = inputs[input_name]

        result = fn(context, **kwargs)

        if inspect.isgenerator(result):
            for item in result:
                yield item
        else:
            if isinstance(result, Result):
                yield result
            elif isinstance(result, MultipleResults):
                for item in result.results:
                    yield item
            elif len(output_defs) == 1:
                yield Result(value=result, output_name=output_defs[0].name)
            elif result is not None:
                if not output_defs:
                    raise DagsterInvariantViolationError(
                        (
                            'Solid unexpectedly returned output {result} of type {type_}. Solid is '
                            'explicitly defined to return no results.'
                        ).format(result=result, type_=type(result))
                    )
                else:
                    raise DagsterInvariantViolationError(
                        (
                            'Solid unexpectedly returned output {result} of type {type_}. Should '
                            'be a MultipleResults object, or a generator, containing or yielding '
                            '{n_results} results: {{{expected_results}}}.'
                        ).format(
                            result=result,
                            type_=type(result),
                            n_results=len(output_defs),
                            expected_results=', '.join(
                                [
                                    '\'{result_name}\': {runtime_type}'.format(
                                        result_name=output_def.name,
                                        runtime_type=output_def.runtime_type,
                                    )
                                    for output_def in output_defs
                                ]
                            ),
                        )
                    )

    return transform


class FunctionValidationError(Exception):
    TYPES = {'vararg': 1, 'missing_name': 2, 'missing_positional': 3, 'extra': 4}

    def __init__(self, error_type, param=None, missing_names=None, **kwargs):
        super(FunctionValidationError, self).__init__(**kwargs)
        self.error_type = error_type
        self.param = param
        self.missing_names = missing_names


def _validate_transform_fn(solid_name, transform_fn, inputs, expected_positionals=None):
    check.str_param(solid_name, 'solid_name')
    check.callable_param(transform_fn, 'transform_fn')
    check.list_param(inputs, 'inputs', of_type=InputDefinition)
    expected_positionals = check.opt_list_param(
        expected_positionals, 'expected_positionals', of_type=(str, tuple)
    )

    names = set(inp.name for inp in inputs)
    # Currently being super strict about naming. Might be a good idea to relax. Starting strict.
    try:
        _validate_decorated_fn(transform_fn, names, expected_positionals)
    except FunctionValidationError as e:
        if e.error_type == FunctionValidationError.TYPES['vararg']:
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function has positional vararg parameter "
                "'{e.param}'. Transform functions should only have keyword arguments that match "
                "input names and a first positional parameter named 'context'.".format(
                    solid_name=solid_name, e=e
                )
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_name']:
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function has parameter '{e.param}' that is not "
                "one of the solid inputs. Transform functions should only have keyword arguments "
                "that match input names and a first positional parameter named 'context'.".format(
                    solid_name=solid_name, e=e
                )
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_positional']:
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function do not have required positional "
                "parameter '{e.param}'. Transform functions should only have keyword arguments "
                "that match input names and a first positional parameter named 'context'.".format(
                    solid_name=solid_name, e=e
                )
            )
        elif e.error_type == FunctionValidationError.TYPES['extra']:
            undeclared_inputs_printed = ", '".join(e.missing_names)
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function do not have parameter(s) "
                "'{undeclared_inputs_printed}', which are in solid's inputs. Transform functions "
                "should only have keyword arguments that match input names and a first positional "
                "parameter named 'context'.".format(
                    solid_name=solid_name, undeclared_inputs_printed=undeclared_inputs_printed
                )
            )
        else:
            raise e


def _validate_decorated_fn(fn, names, expected_positionals):
    used_inputs = set()
    has_kwargs = False

    signature = funcsigs.signature(fn)
    params = list(signature.parameters.values())

    expected_positional_params = params[0 : len(expected_positionals)]
    other_params = params[len(expected_positionals) :]

    for expected_names, actual in zip(expected_positionals, expected_positional_params):
        possible_names = []
        for expected in expected_names:
            possible_names.extend(
                [
                    '_',
                    expected,
                    '_{expected}'.format(expected=expected),
                    '{expected}_'.format(expected=expected),
                ]
            )
        if (
            actual.kind
            not in [funcsigs.Parameter.POSITIONAL_OR_KEYWORD, funcsigs.Parameter.POSITIONAL_ONLY]
        ) or (actual.name not in possible_names):
            raise FunctionValidationError(
                FunctionValidationError.TYPES['missing_positional'], param=expected_names
            )

    for param in other_params:
        if param.kind == funcsigs.Parameter.VAR_KEYWORD:
            has_kwargs = True
        elif param.kind == funcsigs.Parameter.VAR_POSITIONAL:
            raise FunctionValidationError(error_type=FunctionValidationError.TYPES['vararg'])

        else:
            if param.name not in names:
                raise FunctionValidationError(
                    FunctionValidationError.TYPES['missing_name'], param=param.name
                )
            else:
                used_inputs.add(param.name)

    undeclared_inputs = names - used_inputs
    if not has_kwargs and undeclared_inputs:
        raise FunctionValidationError(
            FunctionValidationError.TYPES['extra'], missing_names=undeclared_inputs
        )



from collections import namedtuple
from enum import Enum
import json

from dagster import check

from dagster.utils import merge_dicts
from dagster.utils.error import serializable_error_info_from_exc_info, SerializableErrorInfo

from dagster.utils.logging import (
    DEBUG,
    StructuredLoggerHandler,
    JsonEventLoggerHandler,
    StructuredLoggerMessage,
    check_valid_level_param,
    construct_single_handler_logger,
)

from .log import DagsterLog


class EventType(Enum):
    PIPELINE_START = 'PIPELINE_START'
    PIPELINE_SUCCESS = 'PIPELINE_SUCCESS'
    PIPELINE_FAILURE = 'PIPELINE_FAILURE'

    PIPELINE_PROCESS_START = 'PIPELINE_PROCESS_START'
    PIPELINE_PROCESS_STARTED = 'PIPELINE_PROCESS_STARTED'

    EXECUTION_PLAN_STEP_SUCCESS = 'EXECUTION_PLAN_STEP_SUCCESS'
    EXECUTION_PLAN_STEP_START = 'EXECUTION_PLAN_STEP_START'
    EXECUTION_PLAN_STEP_FAILURE = 'EXECUTION_PLAN_STEP_FAILURE'

    STEP_MATERIALIZATION = 'STEP_MATERIALIZATION'

    UNCATEGORIZED = 'UNCATEGORIZED'


class ExecutionEvents(namedtuple('_ExecutionEvents', 'pipeline_name log')):
    def __new__(cls, pipeline_name, log):
        return super(ExecutionEvents, cls).__new__(
            cls,
            check.str_param(pipeline_name, 'pipeline_name'),
            check.inst_param(log, 'log', DagsterLog),
        )

    def pipeline_start(self):
        self.log.info(
            'Beginning execution of pipeline {pipeline}'.format(pipeline=self.pipeline_name),
            event_type=EventType.PIPELINE_START.value,
        )

    def pipeline_success(self):
        self.log.info(
            'Completing successful execution of pipeline {pipeline}'.format(
                pipeline=self.pipeline_name
            ),
            event_type=EventType.PIPELINE_SUCCESS.value,
        )

    def pipeline_failure(self):
        self.log.info(
            'Completing failing execution of pipeline {pipeline}'.format(
                pipeline=self.pipeline_name
            ),
            event_type=EventType.PIPELINE_FAILURE.value,
        )

    def execution_plan_step_start(self, step_key):
        check.str_param(step_key, 'step_key')
        self.log.info(
            'Beginning execution of {step_key}'.format(step_key=step_key),
            event_type=EventType.EXECUTION_PLAN_STEP_START.value,
            step_key=step_key,
        )

    def execution_plan_step_success(self, step_key, millis):
        check.str_param(step_key, 'step_key')
        check.float_param(millis, 'millis')

        self.log.info(
            'Execution of {step_key} succeeded in {millis}'.format(
                step_key=step_key, millis=millis
            ),
            event_type=EventType.EXECUTION_PLAN_STEP_SUCCESS.value,
            millis=millis,
            step_key=step_key,
        )

    def execution_plan_step_failure(self, step_key, exc_info):
        check.str_param(step_key, 'step_key')
        self.log.info(
            'Execution of {step_key} failed'.format(step_key=step_key),
            event_type=EventType.EXECUTION_PLAN_STEP_FAILURE.value,
            step_key=step_key,
            # We really need a better serialization story here
            error_info=json.dumps(serializable_error_info_from_exc_info(exc_info), sort_keys=True),
        )

    def step_materialization(self, step_key, file_name, file_location):
        check.str_param(step_key, 'step_key')
        self.log.info(
            'Step {step_key} produced materialization of {name} at {loc}'.format(
                step_key=step_key, name=file_name, loc=file_location
            ),
            event_type=EventType.STEP_MATERIALIZATION.value,
            step_key=step_key,
            file_name=file_name,
            file_location=file_location,
        )


EVENT_TYPE_VALUES = {item.value for item in EventType}


def valid_event_type(event_type):
    check.opt_str_param(event_type, 'event_type')
    return event_type in EVENT_TYPE_VALUES


def construct_event_type(event_type):
    check.opt_str_param(event_type, 'event_type')
    return EventType(event_type) if valid_event_type(event_type) else EventType.UNCATEGORIZED


class EventRecord(object):
    def __init__(
        self, error_info, message, level, user_message, event_type, run_id, timestamp, step_key=None
    ):
        self._error_info = check.opt_inst_param(error_info, 'error_info', SerializableErrorInfo)
        self._message = check.str_param(message, 'message')
        self._level = check_valid_level_param(level)
        self._user_message = check.str_param(user_message, 'user_message')
        self._event_type = check.inst_param(event_type, 'event_type', EventType)
        self._run_id = check.str_param(run_id, 'run_id')
        self._timestamp = check.float_param(timestamp, 'timestamp')
        self._step_key = check.opt_str_param(step_key, 'step_key')

    @property
    def message(self):
        return self._message

    @property
    def level(self):
        return self._level

    @property
    def user_message(self):
        return self._user_message

    @property
    def event_type(self):
        return self._event_type

    @property
    def run_id(self):
        return self._run_id

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def error_info(self):
        return self._error_info

    @property
    def step_key(self):
        return self._step_key

    def to_dict(self):
        return {
            'run_id': self.run_id,
            'message': self.message,
            'level': self.level,
            'user_message': self.user_message,
            'event_type': self.event_type.value,
            'timestamp': self.timestamp,
            'error_info': self._error_info,
            'step_key': self._step_key,
        }


class PipelineEventRecord(EventRecord):
    def __init__(self, pipeline_name, **kwargs):
        super(PipelineEventRecord, self).__init__(**kwargs)
        self._pipeline_name = check.str_param(pipeline_name, 'pipeline_name')

    @property
    def pipeline_name(self):
        return self._pipeline_name

    def to_dict(self):
        orig = super(PipelineEventRecord, self).to_dict()
        orig['pipeline_name'] = self.pipeline_name
        return orig


class ExecutionStepEventRecord(EventRecord):
    def __init__(self, pipeline_name, solid_name, solid_definition_name, **kwargs):
        super(ExecutionStepEventRecord, self).__init__(**kwargs)
        self._pipeline_name = check.str_param(pipeline_name, 'pipeline_name')
        self._solid_name = check.str_param(solid_name, 'solid_name')
        self._solid_definition_name = check.str_param(
            solid_definition_name, 'solid_definition_name'
        )

    @property
    def pipeline_name(self):
        return self._pipeline_name

    @property
    def solid_name(self):
        return self._solid_name

    @property
    def solid_definition_name(self):
        return self._solid_definition_name

    def to_dict(self):
        orig = super(ExecutionStepEventRecord, self).to_dict()
        orig.update(
            {
                'pipeline_name': self.pipeline_name,
                'step_key': self.step_key,
                'solid_name': self.solid_name,
                'solid_definition_name': self.solid_definition_name,
            }
        )
        return orig


class ExecutionStepSuccessRecord(ExecutionStepEventRecord):
    def __init__(self, millis, **kwargs):
        super(ExecutionStepSuccessRecord, self).__init__(**kwargs)
        self._millis = check.float_param(millis, 'millis')

    @property
    def millis(self):
        return self._millis

    def to_dict(self):
        orig = super(ExecutionStepSuccessRecord, self).to_dict()
        orig['millis'] = self.millis
        return orig


class LogMessageRecord(EventRecord):
    pass


class StepMaterializationRecord(ExecutionStepEventRecord):
    def __init__(self, file_name, file_location, **kwargs):
        super(StepMaterializationRecord, self).__init__(**kwargs)
        self._name = check.str_param(file_name, 'file_name')
        self._loc = check.str_param(file_location, 'file_location')

    @property
    def file_name(self):
        return self._name

    @property
    def file_location(self):
        return self._loc

    def to_dict(self):
        orig = super(StepMaterializationRecord, self).to_dict()
        orig['name'] = self.file_name
        orig['loc'] = self.file_location
        return orig


EVENT_CLS_LOOKUP = {
    EventType.EXECUTION_PLAN_STEP_FAILURE: ExecutionStepEventRecord,
    EventType.EXECUTION_PLAN_STEP_START: ExecutionStepEventRecord,
    EventType.EXECUTION_PLAN_STEP_SUCCESS: ExecutionStepSuccessRecord,
    EventType.PIPELINE_FAILURE: PipelineEventRecord,
    EventType.PIPELINE_START: PipelineEventRecord,
    EventType.PIPELINE_SUCCESS: PipelineEventRecord,
    EventType.UNCATEGORIZED: LogMessageRecord,
    EventType.STEP_MATERIALIZATION: StepMaterializationRecord,
}

PIPELINE_EVENTS = {EventType.PIPELINE_FAILURE, EventType.PIPELINE_START, EventType.PIPELINE_SUCCESS}


def construct_error_info(logger_message):
    if 'error_info' not in logger_message.meta:
        return None

    raw_error_info = logger_message.meta['error_info']

    message, stack = json.loads(raw_error_info)

    return SerializableErrorInfo(message, stack)


def is_pipeline_event_type(event_type):
    event_cls = EVENT_CLS_LOOKUP[event_type]
    return issubclass(event_cls, PipelineEventRecord)


def logger_to_kwargs(logger_message):
    event_type = construct_event_type(logger_message.meta.get('event_type'))

    base_args = {
        'message': logger_message.message,
        'level': logger_message.level,
        'user_message': logger_message.meta['orig_message'],
        'event_type': event_type,
        'run_id': logger_message.meta['run_id'],
        'timestamp': logger_message.record.created,
        'step_key': logger_message.meta.get('step_key'),
    }

    event_cls = EVENT_CLS_LOOKUP[event_type]
    if issubclass(event_cls, PipelineEventRecord):
        return dict(base_args, pipeline_name=logger_message.meta['pipeline'])
    elif issubclass(event_cls, ExecutionStepEventRecord):
        step_args = {
            'pipeline_name': logger_message.meta['pipeline'],
            'solid_name': logger_message.meta['solid'],
            'solid_definition_name': logger_message.meta['solid_definition'],
        }
        if event_cls == ExecutionStepSuccessRecord:
            step_args['millis'] = logger_message.meta['millis']
        if event_cls == StepMaterializationRecord:
            step_args['file_name'] = logger_message.meta['file_name']
            step_args['file_location'] = logger_message.meta['file_location']

        return merge_dicts(base_args, step_args)
    else:
        return base_args


def construct_event_record(logger_message):
    check.inst_param(logger_message, 'logger_message', StructuredLoggerMessage)
    event_type = construct_event_type(logger_message.meta.get('event_type'))
    log_record_cls = EVENT_CLS_LOOKUP[event_type]

    kwargs = logger_to_kwargs(logger_message)
    return log_record_cls(error_info=construct_error_info(logger_message), **kwargs)


def construct_event_logger(event_record_callback):
    '''
    Callback receives a stream of event_records
    '''
    check.callable_param(event_record_callback, 'event_record_callback')

    return construct_single_handler_logger(
        'event-logger',
        DEBUG,
        StructuredLoggerHandler(
            lambda logger_message: event_record_callback(construct_event_record(logger_message))
        ),
    )


def construct_json_event_logger(json_path):
    '''Record a stream of event records to json'''
    check.str_param(json_path, 'json_path')
    return construct_single_handler_logger(
        "json-event-record-logger",
        DEBUG,
        JsonEventLoggerHandler(
            json_path,
            lambda record: construct_event_record(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,
                    record=record,
                )
            ),
        ),
    )



from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    SolidInstance,
    lambda_solid,
    Int,
)


@lambda_solid(
    inputs=[InputDefinition('num1', Int), InputDefinition('num2', Int)],
    output=OutputDefinition(Int),
)
def adder(num1, num2):
    return num1 + num2


@lambda_solid(
    inputs=[InputDefinition('num1', Int), InputDefinition('num2', Int)],
    output=OutputDefinition(Int),
)
def multer(num1, num2):
    return num1 * num2


def define_reusable_solids_pipeline():
    # (a + b) * (c + d)

    return PipelineDefinition(
        name='reusable_solids_pipeline',
        solids=[adder, multer],
        dependencies={
            SolidInstance('adder', 'a_plus_b'): {},
            SolidInstance('adder', 'c_plus_d'): {},
            SolidInstance('multer', 'final'): {
                'num1': DependencyDefinition('a_plus_b'),
                'num2': DependencyDefinition('c_plus_d'),
            },
        },
    )



from collections import namedtuple

import warnings

from dagster import check

from .definitions.pipeline import PipelineDefinition
from .definitions.resource import ResourceDefinition


class InitContext(namedtuple('_InitContext', 'context_config pipeline_def run_id')):
    '''
    InitContext is the context object context creation functions In effect, it is 
    the state available to those functions, and any function that is called 
    prior to pipeline execution, plus the configuration value for that context.
    '''

    def __new__(cls, context_config, pipeline_def, run_id):
        return super(InitContext, cls).__new__(
            cls,
            context_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.str_param(run_id, 'run_id'),
        )

    @property
    def config(self):
        warnings.warn(
            (
                'As of 0.3.2 the config property is deprecated. Use context_config instead. '
                'This will be removed in 0.4.0.'
            )
        )
        return self.context_config


class InitResourceContext(
    namedtuple(
        'InitResourceContext', 'context_config resource_config pipeline_def resource_def run_id'
    )
):
    '''
    Similar to InitContext, but is resource specific. It includes all the properties
    in the InitContext, plus the resource config and the resource definition
    '''

    def __new__(cls, context_config, resource_config, pipeline_def, resource_def, run_id):
        return super(InitResourceContext, cls).__new__(
            cls,
            context_config,
            resource_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.inst_param(resource_def, 'resource_def', ResourceDefinition),
            check.str_param(run_id, 'run_id'),
        )

    @property
    def config(self):
        warnings.warn(
            (
                'As of 0.3.2 the config property is deprecated. Use resource_config instead.'
                'This will be removed in 0.4.0.'
            )
        )
        return self.resource_config



from collections import defaultdict

from dagster import (
    ExecutionContext,
    InProcessExecutorConfig,
    RunConfig,
    PipelineDefinition,
    PipelineContextDefinition,
    execute_pipeline,
    lambda_solid,
)

from dagster.core.events import construct_event_logger, EventRecord, EventType

from dagster.utils.logging import define_colored_console_logger


def single_event(events, event_type):
    assert event_type in events
    assert len(events[event_type]) == 1
    return events[event_type][0]


def define_event_logging_pipeline(name, solids, event_callback, deps=None):
    return PipelineDefinition(
        name=name,
        solids=solids,
        description=deps,
        context_definitions={
            'default': PipelineContextDefinition(
                context_fn=lambda _: ExecutionContext(
                    loggers=[
                        construct_event_logger(event_callback),
                        define_colored_console_logger('yup'),
                    ]
                )
            )
        },
    )


def test_empty_pipeline():
    events = defaultdict(list)

    def _event_callback(record):
        assert isinstance(record, EventRecord)
        events[record.event_type].append(record)

    pipeline_def = define_event_logging_pipeline(
        name='empty_pipeline', solids=[], event_callback=_event_callback
    )

    result = execute_pipeline(pipeline_def)
    assert result.success
    assert events

    assert single_event(events, EventType.PIPELINE_START).pipeline_name == 'empty_pipeline'
    assert single_event(events, EventType.PIPELINE_SUCCESS).pipeline_name == 'empty_pipeline'


def test_single_solid_pipeline_success():
    events = defaultdict(list)

    @lambda_solid
    def solid_one():
        return 1

    def _event_callback(record):
        events[record.event_type].append(record)

    pipeline_def = define_event_logging_pipeline(
        name='single_solid_pipeline', solids=[solid_one], event_callback=_event_callback
    )

    result = execute_pipeline(pipeline_def)
    assert result.success
    assert events

    start_event = single_event(events, EventType.EXECUTION_PLAN_STEP_START)
    assert start_event.pipeline_name == 'single_solid_pipeline'
    assert start_event.solid_name == 'solid_one'

    assert start_event.solid_definition_name == 'solid_one'
    success_event = single_event(events, EventType.EXECUTION_PLAN_STEP_SUCCESS)
    assert success_event.pipeline_name == 'single_solid_pipeline'
    assert success_event.solid_name == 'solid_one'
    assert success_event.solid_definition_name == 'solid_one'

    assert isinstance(success_event.millis, float)
    assert success_event.millis > 0.0


def test_single_solid_pipeline_failure():
    events = defaultdict(list)

    @lambda_solid
    def solid_one():
        raise Exception('nope')

    def _event_callback(record):
        events[record.event_type].append(record)

    pipeline_def = define_event_logging_pipeline(
        name='single_solid_pipeline', solids=[solid_one], event_callback=_event_callback
    )

    result = execute_pipeline(
        pipeline_def,
        run_config=RunConfig(executor_config=InProcessExecutorConfig(throw_on_user_error=False)),
    )
    assert not result.success

    start_event = single_event(events, EventType.EXECUTION_PLAN_STEP_START)
    assert start_event.pipeline_name == 'single_solid_pipeline'
    assert start_event.solid_name == 'solid_one'
    assert start_event.solid_definition_name == 'solid_one'

    failure_event = single_event(events, EventType.EXECUTION_PLAN_STEP_FAILURE)
    assert failure_event.pipeline_name == 'single_solid_pipeline'
    assert failure_event.solid_name == 'solid_one'
    assert failure_event.solid_definition_name == 'solid_one'



from abc import ABCMeta, abstractmethod
from collections import namedtuple
import pickle

import six

from dagster import check

from dagster.core.files import FileStore


class StepOutputHandle(namedtuple('_StepOutputHandle', 'step_key output_name')):
    @staticmethod
    def from_step(step, output_name):
        from .objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)

        return StepOutputHandle(step.key, output_name)

    def __new__(cls, step_key, output_name):
        return super(StepOutputHandle, cls).__new__(
            cls,
            step_key=check.str_param(step_key, 'step_key'),
            output_name=check.str_param(output_name, 'output_name'),
        )


def read_pickle_file(path):
    check.str_param(path, 'path')
    with open(path, 'rb') as ff:
        return pickle.load(ff)


def write_pickle_file(path, value):
    check.str_param(path, 'path')
    with open(path, 'wb') as ff:
        return pickle.dump(value, ff)


class IntermediatesManager(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def get_value(self, step_output_handle):
        pass

    @abstractmethod
    def set_value(self, step_output_handle, value):
        pass

    @abstractmethod
    def has_value(self, step_output_handle):
        pass

    def all_inputs_covered(self, step):
        from .objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)
        for step_input in step.step_inputs:
            if not self.has_value(step_input.prev_output_handle):
                return False
        return True


class InMemoryIntermediatesManager(IntermediatesManager):
    def __init__(self):
        self.values = {}

    def get_value(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        return self.values[step_output_handle]

    def set_value(self, step_output_handle, value):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        self.values[step_output_handle] = value

    def has_value(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        return step_output_handle in self.values


# TODO: This should go through persistence and serialization infrastructure
# Fixing this requires getting the type information (serialization_strategy is
# a property of runtime type) of the step output you are dealing with. As things
# are structured now we construct the inputs before execution. The fix to this
# will likely be to inject a different type of execution step that threads
# the manager
class FileSystemIntermediateManager(IntermediatesManager):
    def __init__(self, files):
        self._files = check.inst_param(files, 'files', FileStore)

    def _get_path_comps(self, step_output_handle):
        return ['intermediates', step_output_handle.step_key, step_output_handle.output_name]

    def get_value(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        check.invariant(self.has_value(step_output_handle))

        with self._files.readable_binary_stream(*self._get_path_comps(step_output_handle)) as ff:
            return pickle.load(ff)

    def set_value(self, step_output_handle, value):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        check.invariant(not self.has_value(step_output_handle))

        with self._files.writeable_binary_stream(*self._get_path_comps(step_output_handle)) as ff:
            pickle.dump(value, ff)

    def has_value(self, step_output_handle):
        return self._files.has_file(*self._get_path_comps(step_output_handle))



from dagster import check

from .utils import check_valid_name


class ExpectationResult(object):
    ''' Result of an expectation callback.

    When Expectations are evaluated in the callback passed to ExpectationDefinitions,
    the user must return an ExpectationResult object from the callback.

    Attributes:

        success (bool): Whether the expectation passed or not.
        message (str): Information about the computation. Typically only used in the failure case.
        result_context (Any): Arbitrary information about the expectation result.
    '''

    def __init__(self, success, message=None, result_context=None):
        self.success = check.bool_param(success, 'success')
        self.message = check.opt_str_param(message, 'message')
        self.result_context = check.opt_dict_param(result_context, 'result_context')


class ExpectationDefinition(object):
    '''
    Expectations represent a data quality test. It performs an arbitrary computation
    to see if a given input or output satisfies the expectation.

    Attributes:

        name (str): The name of the expectation. Names should be unique per-solid.
        expectation_fn (callable):
            This is the implementation of an expectation computation. It should be a callback
            with the signature (**context**: `ExecutionContext`, **info**:
            `ExpectationExecutionInfo`, **value**: `Any`) : `ExpectationResult`.

            "value" conforms to the type check performed within the Dagster type system.

            e.g. If the expectation is declared on an input of type ``dagster_pandas.DataFrame``,
            you can assume that value is a ``pandas.DataFrame``.

        description (str): Description of expectation. Optional.

    Examples:

        .. code-block:: python

            InputDefinition('some_input', types.Int, expectations=[
                ExpectationDefinition(
                    name='is_positive',
                    expectation_fn=lambda(
                        _info,
                        value,
                    ): ExpectationResult(success=value > 0),
                )
            ])
    '''

    def __init__(self, name, expectation_fn, description=None):
        self.name = check_valid_name(name)
        self.expectation_fn = check.callable_param(expectation_fn, 'expectation_fn')
        self.description = check.opt_str_param(description, 'description')



from dagster.core.execution import create_execution_plan
from .test_multiprocessing import define_diamond_pipeline


def test_topological_sort():
    plan = create_execution_plan(define_diamond_pipeline())

    levels = plan.topological_step_levels()

    assert len(levels) == 3

    assert [step.key for step in levels[0]] == ['return_two.transform']
    assert [step.key for step in levels[1]] == ['add_three.transform', 'mult_three.transform']
    assert [step.key for step in levels[2]] == ['adder.transform']
